"""
Canon CR2 raw image data, version 2.0 image metadata extractor.

Authors: Fernando Crespo
Creation date: 21 february 2017
"""

from hachoir.metadata.metadata import (registerExtractor, RootMetadata)
from hachoir.parser.image import CR2File
from hachoir.metadata.safe import fault_tolerant


class CR2Metadata(RootMetadata):
    key_to_attr = {
        "ImageWidth": "width",
        "ImageLength": "height",
        "ImageDescription": "comment",
        "DocumentName": "title",
        "XResolution": "width_dpi",
        "YResolution": "height_dpi",
    }

    def extract(self, tiff):
        if "ifd[0]" in tiff:
            self.useIFD(tiff["ifd[0]"])
            self.camera_manufacturer = tiff["ifd[0]"]["value[4]"].value
            self.camera_model = tiff["ifd[0]"]["value[5]"].value

        if "exif[0]" in tiff:
            self.date_time_original = tiff["exif[0]"]["value[7]"].value
            self.date_time_digitized = tiff["exif[0]"]["value[8]"].value

    def useIFD(self, ifd):
        attr = {}
        for entry in ifd.array("entry"):
            self.processIfdEntry(ifd, entry, attr)
        if 'BitsPerSample' in attr and 'SamplesPerPixel' in attr:
            self.bits_per_pixel = attr[
                'BitsPerSample'] * attr['SamplesPerPixel']

    @fault_tolerant
    def processIfdEntry(self, ifd, entry, attr):
        tag = entry["tag"].display
        if tag in {"BitsPerSample", "SamplesPerPixel"}:
            value = ifd.getEntryValues(entry)[0].value
            attr[tag] = value
            return

        try:
            attrname = self.key_to_attr[tag]
        except KeyError:
            return
        value = ifd.getEntryValues(entry)[0].value
        if tag in {"XResolution", "YResolution"}:
            value = round(value)
        setattr(self, attrname, value)


registerExtractor(CR2File, CR2Metadata)



from hachoir.field import FieldSet, RawBytes
from hachoir.stream import StringInputStream


class FragmentGroup:

    def __init__(self, parser):
        self.items = []
        self.parser = parser
        self.args = {}

    def add(self, item):
        self.items.append(item)

    def createInputStream(self):
        # FIXME: Use lazy stream creation
        data = []
        for item in self.items:
            data.append(item["rawdata"].value)
        data = b"".join(data)

        tags = {"args": self.args}
        if self.parser is not None:
            tags["class"] = self.parser
        tags = iter(tags.items())
        return StringInputStream(data, "<fragment group>", tags=tags)


class CustomFragment(FieldSet):

    def __init__(self, parent, name, size, parser, description=None, group=None):
        FieldSet.__init__(self, parent, name, description, size=size)
        if not group:
            group = FragmentGroup(parser)
        self.field_size = size
        self.group = group
        self.group.add(self)

    def createFields(self):
        yield RawBytes(self, "rawdata", self.field_size // 8)

    def _createInputStream(self, **args):
        return self.group.createInputStream()

    def createValue(self):
        return self["rawdata"].value

    def createDisplay(self):
        return self["rawdata"].display



from hachoir.field import Bits, Bytes
from hachoir.core.tools import makePrintable, humanFilesize
from hachoir.core import config


class PaddingBits(Bits):
    """
    Padding bits used, for example, to align address (of next field).
    See also NullBits and PaddingBytes types.

    Arguments:
     * nbits: Size of the field in bits

    Optional arguments:
     * pattern (int): Content pattern, eg. 0 if all bits are set to 0
    """
    static_size = staticmethod(lambda *args, **kw: args[1])
    MAX_SIZE = 128

    def __init__(self, parent, name, nbits, description="Padding", pattern=None):
        Bits.__init__(self, parent, name, nbits, description)
        self.pattern = pattern
        self._display_pattern = self.checkPattern()

    def checkPattern(self):
        if not (config.check_padding_pattern):
            return False
        if self.pattern != 0:
            return False

        if self.MAX_SIZE < self._size:
            value = self._parent.stream.readBits(
                self.absolute_address, self.MAX_SIZE, self._parent.endian)
        else:
            value = self.value
        if value != 0:
            self.warning(
                "padding contents doesn't look normal (invalid pattern)")
            return False
        if self.MAX_SIZE < self._size:
            self.info("only check first %u bits" % self.MAX_SIZE)
        return True

    def createDisplay(self):
        if self._display_pattern:
            return "<padding pattern=%s>" % self.pattern
        else:
            return Bits.createDisplay(self)


class PaddingBytes(Bytes):
    """
    Padding bytes used, for example, to align address (of next field).
    See also NullBytes and PaddingBits types.

    Arguments:
     * nbytes: Size of the field in bytes

    Optional arguments:
     * pattern (str): Content pattern, eg. b"\0" for nul bytes
    """

    static_size = staticmethod(lambda *args, **kw: args[1] * 8)
    MAX_SIZE = 4096

    def __init__(self, parent, name, nbytes,
                 description="Padding", pattern=None):
        """ pattern is None or repeated string """
        assert (pattern is None) or (isinstance(pattern, bytes))
        Bytes.__init__(self, parent, name, nbytes, description)
        self.pattern = pattern
        self._display_pattern = self.checkPattern()

    def checkPattern(self):
        if not (config.check_padding_pattern):
            return False
        if self.pattern is None:
            return False

        if self.MAX_SIZE < self._size // 8:
            self.info("only check first %s of padding" %
                      humanFilesize(self.MAX_SIZE))
            content = self._parent.stream.readBytes(
                self.absolute_address, self.MAX_SIZE)
        else:
            content = self.value
        index = 0
        pattern_len = len(self.pattern)
        while index < len(content):
            if content[index:index + pattern_len] != self.pattern:
                self.warning(
                    "padding contents doesn't look normal"
                    " (invalid pattern at byte %u)!"
                    % index)
                return False
            index += pattern_len
        return True

    def createDisplay(self):
        if self._display_pattern:
            return "<padding pattern=%s>" % makePrintable(self.pattern, "ASCII", quote="'")
        else:
            return Bytes.createDisplay(self)

    def createRawDisplay(self):
        return Bytes.createDisplay(self)


class NullBits(PaddingBits):
    """
    Null padding bits used, for example, to align address (of next field).
    See also PaddingBits and NullBytes types.

    Arguments:
     * nbits: Size of the field in bits
    """

    def __init__(self, parent, name, nbits, description=None):
        PaddingBits.__init__(self, parent, name, nbits, description, pattern=0)

    def createDisplay(self):
        if self._display_pattern:
            return "<null>"
        else:
            return Bits.createDisplay(self)


class NullBytes(PaddingBytes):
    """
    Null padding bytes used, for example, to align address (of next field).
    See also PaddingBytes and NullBits types.

    Arguments:
     * nbytes: Size of the field in bytes
    """

    def __init__(self, parent, name, nbytes, description=None):
        PaddingBytes.__init__(self, parent, name, nbytes,
                              description, pattern=b"\0")

    def createDisplay(self):
        if self._display_pattern:
            return "<null>"
        else:
            return Bytes.createDisplay(self)



from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN
from hachoir.field import RawBytes, RootSeekableFieldSet, ParserError
from hachoir.parser import HachoirParser


class OLE2FragmentParser(HachoirParser, RootSeekableFieldSet):
    tags = {
        "description": "Microsoft Office document subfragments",
    }
    endian = LITTLE_ENDIAN

    ENDIAN_CHECK = False

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)
        if self.ENDIAN_CHECK:
            if self["endian"].value == b"\xFF\xFE":
                self.endian = BIG_ENDIAN
            elif self["endian"].value == b"\xFE\xFF":
                self.endian = LITTLE_ENDIAN
            else:
                raise ParserError("OLE2: Invalid endian value")

    def validate(self):
        if self.ENDIAN_CHECK:
            if self["endian"].value not in ["\xFF\xFE", "\xFE\xFF"]:
                return "Unknown endian value %s" % self["endian"].value.encode('hex')
        return True


class RawParser(OLE2FragmentParser):
    ENDIAN_CHECK = False
    OS_CHECK = False

    def createFields(self):
        yield RawBytes(self, "rawdata", self.datasize)
        if self.datasize < self.size // 8:
            yield RawBytes(self, "slack_space", (self.size // 8) - self.datasize)



from hachoir.stream import FileOutputStream, StreamError
from hachoir.core.error import error
from errno import EEXIST
from os import mkdir, path


class Output:
    """
    Store files found by search tool.
    """

    def __init__(self, directory):
        self.directory = directory
        self.mkdir = False
        self.file_id = 1

    def createDirectory(self):
        try:
            mkdir(self.directory)
        except OSError as err:
            if err.errno == EEXIST:
                pass
            else:
                raise

    def createFilename(self, file_ext=None):
        filename = "file-%04u" % self.file_id
        self.file_id += 1
        if file_ext:
            filename += file_ext
        return filename

    def writeFile(self, filename, stream, offset, size):
        # Create directory (only on first call)
        if not self.mkdir:
            self.createDirectory()
            self.mkdir = True

        # Create output file
        filename = path.join(self.directory, filename)
        output = FileOutputStream(filename)

        # Write output
        try:
            output.copyBytesFrom(stream, offset, size // 8)
        except StreamError as err:
            error("copyBytesFrom() error: %s" % err)
        return filename



from wx import App, EVT_MENU, ID_OK
from wx.xrc import XRCID

import hachoir.core.config as config
from hachoir.parser.guess import createParser, guessParser
from hachoir.stream.input import FileFromInputStream
from hachoir.wx.dispatcher import dispatcher_t
from hachoir.wx import frame_view, field_view, hex_view, tree_view
from hachoir.wx.dialogs import file_open_dialog
from hachoir.wx.unicode import force_unicode
from hachoir import __version__


class app_t(App):
    def __init__(self, filename=None):
        print("[+] Run hachoir-wx version %s" % __version__)
        self.filename = filename
        App.__init__(self, False)

    def OnInit(self):
        self.bind_events()

        if self.filename:
            load_file(self, self.filename)
        else:
            self.on_file_menu_open_file(None)

        return True

    def bind_events(self):
        self.Bind(EVT_MENU, self.on_file_menu_open_file,
                  id=XRCID('file_menu_open_file'))
        self.Bind(EVT_MENU, self.on_file_menu_close_window,
                  id=XRCID('file_menu_close_window'))

    def on_file_menu_open_file(self, event):
        open_dialog = file_open_dialog()
        if ID_OK != open_dialog.ShowModal():
            return
        filename = open_dialog.GetPath()
        open_dialog.Destroy()
        load_file(self, filename)

    def on_field_parse_substream(self, dispatcher, field):
        stream = field.getSubIStream()
        parser = guessParser(stream)
        if not parser:
            msg = '[-] Unable to parse substream.'
            if not config.verbose:
                msg += ' (use --verbose to see validation errors)'
            print(msg)
            return
        print('[+] Substream parser: %s.%s' % (parser.__module__, type(parser).__name__))
        subfile = FileFromInputStream(stream)
        subfile.name = field.path
        new_window(self, subfile, parser, subfile.name)

    def on_field_open_window_here(self, dispatcher, field):
        new_window(self, dispatcher.top_file, field, dispatcher.top_filename)

    def on_frame_activated(self, dispatcher, frame):
        self.SetTopWindow(frame)

    def on_file_menu_close_window(self, event):
        top_window = self.GetTopWindow()
        assert top_window is not None
        top_window.Close()


def load_file(app, filename):
    parser = createParser(force_unicode(filename), real_filename=filename)
    if not parser:
        return
    print('[+] Using parser: %s.%s' % (parser.__module__, type(parser).__name__))
    new_window(app, open(filename, 'rb'), parser, filename)


def new_window(app, file, parser, filename):
    print('[+] Opening new GUI')

    dispatcher = dispatcher_t()
    dispatcher.add_receiver(app)
    dispatcher.top_file = file
    dispatcher.top_filename = filename

    frame = frame_view.setup_frame_view(dispatcher)
    field_view.setup_field_view(frame, dispatcher)
    hex_view_widget = hex_view.setup_hex_view(frame, dispatcher)
    tree_view.setup_tree_view(frame, dispatcher)

    dispatcher.trigger('file_ready', file)
    dispatcher.trigger('field_set_ready', parser)
    dispatcher.trigger('filename_update', filename.replace('\\', '/'))
    frame.ready()
    hex_view_widget.ready()

    frame.Show()
    print('[+] GUI ready')



from hachoir.field import MissingField
from hachoir.metadata.metadata import (registerExtractor,
                                       Metadata, RootMetadata, MultipleMetadata)
from hachoir.metadata.metadata_item import QUALITY_GOOD
from hachoir.metadata.safe import fault_tolerant
from hachoir.parser.video import AsfFile, FlvFile
from hachoir.parser.video.asf import Descriptor as ASF_Descriptor
from hachoir.parser.container import MkvFile, MP4File
from hachoir.parser.container.mkv import dateToDatetime
from hachoir.core.tools import makeUnicode, makePrintable, timedelta2seconds
from datetime import timedelta


class MkvMetadata(MultipleMetadata):
    tag_key = {
        "TITLE": "title",
        "URL": "url",
        "COPYRIGHT": "copyright",

        # TODO: use maybe another name?
        # Its value may be different than (...)/Info/DateUTC/date
        "DATE_RECORDED": "creation_date",

        # TODO: Extract subtitle metadata
        "SUBTITLE": "subtitle_author",
    }

    def extract(self, mkv):
        for segment in mkv.array("Segment"):
            self.processSegment(segment)

    def processSegment(self, segment):
        for field in segment:
            if field.name.startswith("Info["):
                self.processInfo(field)
            elif field.name.startswith("Tags["):
                for tag in field.array("Tag"):
                    self.processTag(tag)
            elif field.name.startswith("Tracks["):
                self.processTracks(field)
            elif field.name.startswith("Cluster["):
                if self.quality < QUALITY_GOOD:
                    return

    def processTracks(self, tracks):
        for entry in tracks.array("TrackEntry"):
            self.processTrack(entry)

    def processTrack(self, track):
        if "TrackType/enum" not in track:
            return
        if track["TrackType/enum"].display == "video":
            self.processVideo(track)
        elif track["TrackType/enum"].display == "audio":
            self.processAudio(track)
        elif track["TrackType/enum"].display == "subtitle":
            self.processSubtitle(track)

    def trackCommon(self, track, meta):
        if "Name/unicode" in track:
            meta.title = track["Name/unicode"].value
        if "Language/string" in track:
            meta.language = track["Language/string"].value
        else:
            meta.language = "eng"

    def processVideo(self, track):
        video = Metadata(self)
        self.trackCommon(track, video)
        try:
            video.compression = track["CodecID/string"].value
            if "Video" in track:
                video.width = track["Video/PixelWidth/unsigned"].value
                video.height = track["Video/PixelHeight/unsigned"].value
        except MissingField:
            pass
        self.addGroup("video[]", video, "Video stream")

    def getDouble(self, field, parent):
        float_key = '%s/float' % parent
        if float_key in field:
            return field[float_key].value
        double_key = '%s/double' % parent
        if double_key in field:
            return field[double_key].value
        return None

    def processAudio(self, track):
        audio = Metadata(self)
        self.trackCommon(track, audio)
        if "Audio" in track:
            frequency = self.getDouble(track, "Audio/SamplingFrequency")
            if frequency is not None:
                audio.sample_rate = frequency
            if "Audio/Channels/unsigned" in track:
                audio.nb_channel = track["Audio/Channels/unsigned"].value
            if "Audio/BitDepth/unsigned" in track:
                audio.bits_per_sample = track["Audio/BitDepth/unsigned"].value
        if "CodecID/string" in track:
            audio.compression = track["CodecID/string"].value
        self.addGroup("audio[]", audio, "Audio stream")

    def processSubtitle(self, track):
        sub = Metadata(self)
        self.trackCommon(track, sub)
        try:
            sub.compression = track["CodecID/string"].value
        except MissingField:
            pass
        self.addGroup("subtitle[]", sub, "Subtitle")

    def processTag(self, tag):
        for field in tag.array("SimpleTag"):
            self.processSimpleTag(field)

    def processSimpleTag(self, tag):
        if "TagName/unicode" not in tag \
                or "TagString/unicode" not in tag:
            return
        name = tag["TagName/unicode"].value
        if name not in self.tag_key:
            return
        key = self.tag_key[name]
        value = tag["TagString/unicode"].value
        setattr(self, key, value)

    def processInfo(self, info):
        if "TimecodeScale/unsigned" in info:
            duration = self.getDouble(info, "Duration")
            if duration is not None:
                try:
                    seconds = duration * \
                        info["TimecodeScale/unsigned"].value * 1e-9
                    self.duration = timedelta(seconds=seconds)
                except OverflowError:
                    # Catch OverflowError for timedelta (long int too large
                    # to be converted to an int)
                    pass
        if "DateUTC/date" in info:
            try:
                self.creation_date = dateToDatetime(info["DateUTC/date"].value)
            except OverflowError:
                pass
        if "WritingApp/unicode" in info:
            self.producer = info["WritingApp/unicode"].value
        if "MuxingApp/unicode" in info:
            self.producer = info["MuxingApp/unicode"].value
        if "Title/unicode" in info:
            self.title = info["Title/unicode"].value


class FlvMetadata(MultipleMetadata):

    def extract(self, flv):
        if "video[0]" in flv:
            meta = Metadata(self)
            self.extractVideo(flv["video[0]"], meta)
            self.addGroup("video", meta, "Video stream")
        if "audio[0]" in flv:
            meta = Metadata(self)
            self.extractAudio(flv["audio[0]"], meta)
            self.addGroup("audio", meta, "Audio stream")
        # TODO: Computer duration
        # One technic: use last video/audio chunk and use timestamp
        # But this is very slow
        self.format_version = flv.description

        if "metadata/entry[1]" in flv:
            self.extractAMF(flv["metadata/entry[1]"])
        if self.has('duration'):
            self.bit_rate = flv.size / timedelta2seconds(self.get('duration'))

    @fault_tolerant
    def extractAudio(self, audio, meta):
        if audio["codec"].display == "MP3" and "music_data" in audio:
            meta.compression = audio["music_data"].description
        else:
            meta.compression = audio["codec"].display
        meta.sample_rate = audio.getSampleRate()
        if audio["is_16bit"].value:
            meta.bits_per_sample = 16
        else:
            meta.bits_per_sample = 8
        if audio["is_stereo"].value:
            meta.nb_channel = 2
        else:
            meta.nb_channel = 1

    @fault_tolerant
    def extractVideo(self, video, meta):
        meta.compression = video["codec"].display

    def extractAMF(self, amf):
        for entry in amf.array("item"):
            self.useAmfEntry(entry)

    @fault_tolerant
    def useAmfEntry(self, entry):
        key = entry["key"].value
        if key == "duration":
            self.duration = timedelta(seconds=entry["value"].value)
        elif key == "creator":
            self.producer = entry["value"].value
        elif key == "audiosamplerate":
            self.sample_rate = entry["value"].value
        elif key == "framerate":
            self.frame_rate = entry["value"].value
        elif key == "metadatacreator":
            self.producer = entry["value"].value
        elif key == "metadatadate":
            self.creation_date = entry.value
        elif key == "width":
            self.width = int(entry["value"].value)
        elif key == "height":
            self.height = int(entry["value"].value)


class MP4Metadata(RootMetadata):

    def extract(self, mov):
        for atom in mov:
            if "movie" in atom:
                self.processMovie(atom["movie"])

    @fault_tolerant
    def processMovieHeader(self, hdr):
        self.creation_date = hdr["creation_date"].value
        self.last_modification = hdr["lastmod_date"].value
        self.duration = timedelta(seconds=float(
            hdr["duration"].value) / hdr["time_scale"].value)
        self.comment = "Play speed: %.1f%%" % (hdr["play_speed"].value * 100)
        self.comment = "User volume: %.1f%%" % (
            float(hdr["volume"].value) * 100)

    @fault_tolerant
    def processTrackHeader(self, hdr):
        width = int(hdr["frame_size_width"].value)
        height = int(hdr["frame_size_height"].value)
        if width and height:
            self.width = width
            self.height = height

    def processTrack(self, atom):
        for field in atom:
            if "track_hdr" in field:
                self.processTrackHeader(field["track_hdr"])

    def processMovie(self, atom):
        for field in atom:
            if "track" in field:
                self.processTrack(field["track"])
            if "movie_hdr" in field:
                self.processMovieHeader(field["movie_hdr"])


class AsfMetadata(MultipleMetadata):
    EXT_DESC_TO_ATTR = {
        "Encoder": "producer",
        "ToolName": "producer",
        "AlbumTitle": "album",
        "Track": "track_number",
        "TrackNumber": "track_total",
        "Year": "creation_date",
        "AlbumArtist": "author",
    }
    SKIP_EXT_DESC = set((
        # Useless informations
        "WMFSDKNeeded", "WMFSDKVersion",
        "Buffer Average", "VBR Peak", "EncodingTime",
        "MediaPrimaryClassID", "UniqueFileIdentifier",
    ))

    def extract(self, asf):
        if "header/content" in asf:
            self.processHeader(asf["header/content"])

    def processHeader(self, header):
        compression = []
        is_vbr = None

        if "ext_desc/content" in header:
            # Extract all data from ext_desc
            data = {}
            for desc in header.array("ext_desc/content/descriptor"):
                self.useExtDescItem(desc, data)

            # Have ToolName and ToolVersion? If yes, group them to producer key
            if "ToolName" in data and "ToolVersion" in data:
                self.producer = "%s (version %s)" % (
                    data["ToolName"], data["ToolVersion"])
                del data["ToolName"]
                del data["ToolVersion"]

            # "IsVBR" key
            if "IsVBR" in data:
                is_vbr = (data["IsVBR"] == 1)
                del data["IsVBR"]

            # Store data
            for key, value in data.items():
                if key in self.EXT_DESC_TO_ATTR:
                    key = self.EXT_DESC_TO_ATTR[key]
                else:
                    if isinstance(key, str):
                        key = makePrintable(key, "ISO-8859-1")
                    value = "%s=%s" % (key, value)
                    key = "comment"
                setattr(self, key, value)

        if "file_prop/content" in header:
            self.useFileProp(header["file_prop/content"], is_vbr)

        if "codec_list/content" in header:
            for codec in header.array("codec_list/content/codec"):
                if "name" in codec:
                    text = codec["name"].value
                    if "desc" in codec and codec["desc"].value:
                        text = "%s (%s)" % (text, codec["desc"].value)
                    compression.append(text)

        audio_index = 1
        video_index = 1
        for index, stream_prop in enumerate(header.array("stream_prop")):
            if "content/audio_header" in stream_prop:
                meta = Metadata(self)
                self.streamProperty(header, index, meta)
                self.streamAudioHeader(
                    stream_prop["content/audio_header"], meta)
                if self.addGroup("audio[%u]" % audio_index, meta, "Audio stream #%u" % audio_index):
                    audio_index += 1
            elif "content/video_header" in stream_prop:
                meta = Metadata(self)
                self.streamProperty(header, index, meta)
                self.streamVideoHeader(
                    stream_prop["content/video_header"], meta)
                if self.addGroup("video[%u]" % video_index, meta, "Video stream #%u" % video_index):
                    video_index += 1

        if "metadata/content" in header:
            info = header["metadata/content"]
            try:
                self.title = info["title"].value
                self.author = info["author"].value
                self.copyright = info["copyright"].value
            except MissingField:
                pass

    @fault_tolerant
    def streamAudioHeader(self, audio, meta):
        if not meta.has("compression"):
            meta.compression = audio["twocc"].display
        meta.nb_channel = audio["channels"].value
        meta.sample_rate = audio["sample_rate"].value
        meta.bits_per_sample = audio["bits_per_sample"].value

    @fault_tolerant
    def streamVideoHeader(self, video, meta):
        meta.width = video["width"].value
        meta.height = video["height"].value
        if "bmp_info" in video:
            bmp_info = video["bmp_info"]
            if not meta.has("compression"):
                meta.compression = bmp_info["codec"].display
            meta.bits_per_pixel = bmp_info["bpp"].value

    @fault_tolerant
    def useExtDescItem(self, desc, data):
        if desc["type"].value == ASF_Descriptor.TYPE_BYTE_ARRAY:
            # Skip binary data
            return
        key = desc["name"].value
        if "/" in key:
            # Replace "WM/ToolName" with "ToolName"
            key = key.split("/", 1)[1]
        if key in self.SKIP_EXT_DESC:
            # Skip some keys
            return
        value = desc["value"].value
        if not value:
            return
        value = makeUnicode(value)
        data[key] = value

    @fault_tolerant
    def useFileProp(self, prop, is_vbr):
        self.creation_date = prop["creation_date"].value
        self.duration = prop["play_duration"].value - prop["preroll"].value
        if prop["seekable"].value:
            self.comment = "Is seekable"
        value = prop["max_bitrate"].value
        text = prop["max_bitrate"].display
        if is_vbr is True:
            text = "VBR (%s max)" % text
        elif is_vbr is False:
            text = "%s (CBR)" % text
        else:
            text = "%s (max)" % text
        self.bit_rate = (value, text)

    def streamProperty(self, header, index, meta):
        key = "bit_rates/content/bit_rate[%u]/avg_bitrate" % index
        if key in header:
            meta.bit_rate = header[key].value

        # TODO: Use codec list
        # It doesn't work when the video uses /header/content/bitrate_mutex
        # since the codec list are shared between streams but... how is it
        # shared?
#        key = "codec_list/content/codec[%u]" % index
#        if key in header:
#            codec = header[key]
#            if "name" in codec:
#                text = codec["name"].value
#                if "desc" in codec and codec["desc"].value:
#                    meta.compression = "%s (%s)" % (text, codec["desc"].value)
#                else:
#                    meta.compression = text


registerExtractor(MP4File, MP4Metadata)
registerExtractor(AsfFile, AsfMetadata)
registerExtractor(FlvFile, FlvMetadata)
registerExtractor(MkvFile, MkvMetadata)



from hachoir.wx.resource import get_child_control


def setup_hex_view(parent, dispatcher):
    print("[+] Setup hex view")
    hex_view = get_child_control(parent, 'hex_view')
    dispatcher.add_sender(hex_view)
    dispatcher.add(hex_view)
    dispatcher.add_receiver(hex_view)

    return hex_view



from hachoir.core.tools import humanDurationNanosec
from math import floor
from time import time


class BenchmarkError(Exception):
    """
    Error during benchmark, use str(err) to format it as string.
    """

    def __init__(self, message):
        Exception.__init__(self, "Benchmark internal error: %s" % message)


class BenchmarkStat:
    """
    Benchmark statistics. This class automatically computes minimum value,
    maximum value and sum of all values.

    Methods:
    - append(value): append a value
    - getMin(): minimum value
    - getMax(): maximum value
    - getSum(): sum of all values
    - __len__(): get number of elements
    - __nonzero__(): isn't empty?
    """

    def __init__(self):
        self._values = []

    def append(self, value):
        self._values.append(value)
        try:
            self._min = min(self._min, value)
            self._max = max(self._max, value)
            self._sum += value
        except AttributeError:
            self._min = value
            self._max = value
            self._sum = value

    def __len__(self):
        return len(self._values)

    def __bool__(self):
        return bool(self._values)

    def getMin(self):
        return self._min

    def getMax(self):
        return self._max

    def getSum(self):
        return self._sum


class Benchmark:

    def __init__(self, max_time=5.0,
                 min_count=5, max_count=None, progress_time=1.0):
        """
        Constructor:
        - max_time: Maximum wanted duration of the whole benchmark
          (default: 5 seconds, minimum: 1 second).
        - min_count: Minimum number of function calls to get good statistics
          (defaut: 5, minimum: 1).
        - progress_time: Time between each "progress" message
          (default: 1 second, minimum: 250 ms).
        - max_count: Maximum number of function calls (default: no limit).
        - verbose: Is verbose? (default: False)
        - disable_gc: Disable garbage collector? (default: False)
        """
        self.max_time = max(max_time, 1.0)
        self.min_count = max(min_count, 1)
        self.max_count = max_count
        self.progress_time = max(progress_time, 0.25)
        self.verbose = False
        self.disable_gc = False

    def formatTime(self, value):
        """
        Format a time delta to string: use humanDurationNanosec()
        """
        return humanDurationNanosec(value * 1000000000)

    def displayStat(self, stat):
        """
        Display statistics to stdout:
        - best time (minimum)
        - average time (arithmetic average)
        - worst time (maximum)
        - total time (sum)

        Use arithmetic avertage instead of geometric average because
        geometric fails if any value is zero (returns zero) and also
        because floating point multiplication lose precision with many
        values.
        """
        average = stat.getSum() / len(stat)
        values = (stat.getMin(), average, stat.getMax(), stat.getSum())
        values = tuple(self.formatTime(value) for value in values)
        print("Benchmark: best=%s  average=%s  worst=%s  total=%s" % values)

    def _runOnce(self, func, args, kw):
        before = time()
        func(*args, **kw)
        after = time()
        return after - before

    def _run(self, func, args, kw):
        """
        Call func(*args, **kw) as many times as needed to get
        good statistics. Algorithm:
        - call the function once
        - compute needed number of calls
        - and then call function N times

        To compute number of calls, parameters are:
        - time of first function call
        - minimum number of calls (min_count attribute)
        - maximum test time (max_time attribute)

        Notice: The function will approximate number of calls.
        """
        # First call of the benchmark
        stat = BenchmarkStat()
        diff = self._runOnce(func, args, kw)
        best = diff
        stat.append(diff)
        total_time = diff

        # Compute needed number of calls
        count = int(floor(self.max_time / diff))
        count = max(count, self.min_count)
        if self.max_count:
            count = min(count, self.max_count)

        # Not other call? Just exit
        if count == 1:
            return stat
        estimate = diff * count
        if self.verbose:
            print("Run benchmark: %s calls (estimate: %s)"
                  % (count, self.formatTime(estimate)))

        display_progress = self.verbose and (1.0 <= estimate)
        total_count = 1
        while total_count < count:
            # Run benchmark and display each result
            if display_progress:
                print("Result %s/%s: %s  (best: %s)" %
                      (total_count, count,
                       self.formatTime(diff), self.formatTime(best)))
            part = count - total_count

            # Will takes more than one second?
            average = total_time / total_count
            if self.progress_time < part * average:
                part = max(int(self.progress_time / average), 1)
            for index in range(part):
                diff = self._runOnce(func, args, kw)
                stat.append(diff)
                total_time += diff
                best = min(diff, best)
            total_count += part
        if display_progress:
            print("Result %s/%s: %s  (best: %s)" %
                  (count, count,
                   self.formatTime(diff), self.formatTime(best)))
        return stat

    def validateStat(self, stat):
        """
        Check statistics and raise a BenchmarkError if they are invalid.
        Example of tests: reject empty stat, reject stat with only nul values.
        """
        if not stat:
            raise BenchmarkError("empty statistics")
        if not stat.getSum():
            raise BenchmarkError("nul statistics")

    def run(self, func, *args, **kw):
        """
        Run function func(*args, **kw), validate statistics,
        and display the result on stdout.

        Disable garbage collector if asked too.
        """

        # Disable garbarge collector is needed and if it does exist
        # (Jython 2.2 don't have it for example)
        if self.disable_gc:
            try:
                import gc
            except ImportError:
                self.disable_gc = False
        if self.disable_gc:
            gc_enabled = gc.isenabled()
            gc.disable()
        else:
            gc_enabled = False

        # Run the benchmark
        stat = self._run(func, args, kw)
        if gc_enabled:
            gc.enable()

        # Validate and display stats
        self.validateStat(stat)
        self.displayStat(stat)



"""
Microsoft Cabinet (CAB) archive.

Author: Victor Stinner, Robert Xiao
Creation date: 31 january 2007

- Microsoft Cabinet SDK
  http://msdn2.microsoft.com/en-us/library/ms974336.aspx
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, Enum,
                           CString, String,
                           UInt8, UInt16, UInt32, Bit, Bits, PaddingBits, NullBits,
                           DateTimeMSDOS32, RawBytes, CustomFragment)
from hachoir.core.text_handler import textHandler, hexadecimal, filesizeHandler
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.tools import paddingSize
from hachoir.stream import StringInputStream
from hachoir.parser.archive.lzx import LZXStream, lzx_decompress
from hachoir.parser.archive.zlib import DeflateBlock

MAX_NB_FOLDER = 30

COMPRESSION_NONE = 0
COMPRESSION_NAME = {
    0: "Uncompressed",
    1: "Deflate",
    2: "Quantum",
    3: "LZX",
}


class Folder(FieldSet):

    def createFields(self):
        yield UInt32(self, "offset", "Offset to data (from file start)")
        yield UInt16(self, "data_blocks", "Number of data blocks which are in this cabinet")
        yield Enum(Bits(self, "compr_method", 4, "Compression method"), COMPRESSION_NAME)
        if self["compr_method"].value in [2, 3]:  # Quantum or LZX use compression level
            yield PaddingBits(self, "padding[]", 4)
            yield Bits(self, "compr_level", 5, "Compression level")
            yield PaddingBits(self, "padding[]", 3)
        else:
            yield PaddingBits(self, "padding[]", 12)
        if self["../flags/has_reserved"].value and self["../reserved_folder_size"].value:
            yield RawBytes(self, "reserved_folder", self["../reserved_folder_size"].value, "Per-folder reserved area")

    def createDescription(self):
        text = "Folder: compression %s" % self["compr_method"].display
        if self["compr_method"].value in [2, 3]:  # Quantum or LZX use compression level
            text += " (level %u: window size %u)" % (
                self["compr_level"].value, 2**self["compr_level"].value)
        return text


class CabFileAttributes(FieldSet):

    def createFields(self):
        yield Bit(self, "readonly")
        yield Bit(self, "hidden")
        yield Bit(self, "system")
        yield Bits(self, "reserved[]", 2)
        yield Bit(self, "archive", "Has the file been modified since the last backup?")
        yield Bit(self, "exec", "Run file after extraction?")
        yield Bit(self, "name_is_utf", "Is the filename using UTF-8?")
        yield Bits(self, "reserved[]", 8)


class File(FieldSet):

    def createFields(self):
        yield filesizeHandler(UInt32(self, "filesize", "Uncompressed file size"))
        yield UInt32(self, "folder_offset", "File offset in uncompressed folder")
        yield Enum(UInt16(self, "folder_index", "Containing folder ID (index)"), {
            0xFFFD: "Folder continued from previous cabinet (real folder ID = 0)",
            0xFFFE: "Folder continued to next cabinet (real folder ID = %i)" % (self["../nb_folder"].value - 1),
            0xFFFF: "Folder spanning previous, current and next cabinets (real folder ID = 0)"})
        yield DateTimeMSDOS32(self, "timestamp")
        yield CabFileAttributes(self, "attributes")
        if self["attributes/name_is_utf"].value:
            yield CString(self, "filename", charset="UTF-8")
        else:
            yield CString(self, "filename", charset="ASCII")

    def createDescription(self):
        return "File %s (%s)" % (
            self["filename"].display, self["filesize"].display)


class Flags(FieldSet):
    static_size = 16

    def createFields(self):
        yield Bit(self, "has_previous")
        yield Bit(self, "has_next")
        yield Bit(self, "has_reserved")
        yield NullBits(self, "padding", 13)


class DataBlock(FieldSet):

    def __init__(self, *args, **kwargs):
        FieldSet.__init__(self, *args, **kwargs)
        size = (self["size"].value + 8) * 8  # +8 for header values
        if self["/flags/has_reserved"].value:
            size += self["/reserved_data_size"].value * 8
        self._size = size

    def createFields(self):
        yield textHandler(UInt32(self, "crc32"), hexadecimal)
        yield UInt16(self, "size")
        yield UInt16(self, "uncompressed_size", "If this is 0, this block is continued in a subsequent cabinet")
        if self["/flags/has_reserved"].value and self["/reserved_data_size"].value:
            yield RawBytes(self, "reserved_data", self["/reserved_data_size"].value, "Per-datablock reserved area")
        compr_method = self.parent.folder["compr_method"].value
        if compr_method == 0:  # Uncompressed
            yield RawBytes(self, "data", self["size"].value, "Folder Data")
            self.parent.uncompressed_data += self["data"].value
        elif compr_method == 1:  # MSZIP
            yield String(self, "mszip_signature", 2, "MSZIP Signature (CK)")
            yield DeflateBlock(self, "deflate_block", self.parent.uncompressed_data)
            padding = paddingSize(self.current_size, 8)
            if padding:
                yield PaddingBits(self, "padding[]", padding)
            self.parent.uncompressed_data = self["deflate_block"].uncomp_data
        elif compr_method == 2:  # Quantum
            yield RawBytes(self, "compr_data", self["size"].value, "Compressed Folder Data")
        elif compr_method == 3:  # LZX
            group = getattr(self.parent.folder, "lzx_group", None)
            field = CustomFragment(
                self, "data", self["size"].value * 8, LZXStream, "LZX data fragment", group)
            if group is None:
                field.group.args["compr_level"] = self.parent.folder["compr_level"].value
            self.parent.folder.lzx_group = field.group
            yield field


class FolderParser(Parser):
    endian = LITTLE_ENDIAN

    def createFields(self):
        for file in sorted(self.files, key=lambda x: x["folder_offset"].value):
            padding = self.seekByte(file["folder_offset"].value)
            if padding:
                yield padding
            yield RawBytes(self, "file[]", file["filesize"].value, file.description)


class FolderData(FieldSet):

    def __init__(self, parent, name, folder, files, *args, **kwargs):
        FieldSet.__init__(self, parent, name, *args, **kwargs)

        def createInputStream(cis, source=None, **args):
            stream = cis(source=source)
            tags = args.setdefault("tags", [])
            tags.extend(stream.tags)
            tags.append(("class", FolderParser))
            tags.append(("args", {'files': files}))
            for unused in self:
                pass
            if folder["compr_method"].value == 3:  # LZX
                self.uncompressed_data = lzx_decompress(
                    self["block[0]/data"].getSubIStream(), folder["compr_level"].value)
            return StringInputStream(self.uncompressed_data, source=source, **args)
        self.setSubIStream(createInputStream)
        self.files = files
        self.folder = folder  # Folder fieldset

    def createFields(self):
        self.uncompressed_data = ""
        for index in range(self.folder["data_blocks"].value):
            block = DataBlock(self, "block[]")
            for i in block:
                pass
            yield block


class CabFile(Parser):
    endian = LITTLE_ENDIAN
    MAGIC = b"MSCF"
    PARSER_TAGS = {
        "id": "cab",
        "category": "archive",
        "file_ext": ("cab",),
        "mime": ("application/vnd.ms-cab-compressed",),
        "magic": ((MAGIC, 0),),
        "min_size": 1 * 8,  # header + file entry
        "description": "Microsoft Cabinet archive"
    }

    def validate(self):
        if self.stream.readBytes(0, 4) != self.MAGIC:
            return "Invalid magic"
        if self["major_version"].value != 1 or self["minor_version"].value != 3:
            return "Unknown version (%i.%i)" % (self["major_version"].value, self["minor_version"].value)
        if not (1 <= self["nb_folder"].value <= MAX_NB_FOLDER):
            return "Invalid number of folder (%s)" % self["nb_folder"].value
        return True

    def createFields(self):
        yield String(self, "magic", 4, "Magic (MSCF)", charset="ASCII")
        yield textHandler(UInt32(self, "hdr_checksum", "Header checksum (0 if not used)"), hexadecimal)
        yield filesizeHandler(UInt32(self, "filesize", "Cabinet file size"))
        yield textHandler(UInt32(self, "fld_checksum", "Folders checksum (0 if not used)"), hexadecimal)
        yield UInt32(self, "off_file", "Offset of first file")
        yield textHandler(UInt32(self, "files_checksum", "Files checksum (0 if not used)"), hexadecimal)
        yield UInt8(self, "minor_version", "Minor version (should be 3)")
        yield UInt8(self, "major_version", "Major version (should be 1)")
        yield UInt16(self, "nb_folder", "Number of folders")
        yield UInt16(self, "nb_files", "Number of files")
        yield Flags(self, "flags")
        yield UInt16(self, "setid")
        yield UInt16(self, "cabinet_serial", "Zero-based cabinet number")

        if self["flags/has_reserved"].value:
            yield UInt16(self, "reserved_header_size", "Size of per-cabinet reserved area")
            yield UInt8(self, "reserved_folder_size", "Size of per-folder reserved area")
            yield UInt8(self, "reserved_data_size", "Size of per-datablock reserved area")
            if self["reserved_header_size"].value:
                yield RawBytes(self, "reserved_header", self["reserved_header_size"].value, "Per-cabinet reserved area")
        if self["flags/has_previous"].value:
            yield CString(self, "previous_cabinet", "File name of previous cabinet", charset="ASCII")
            yield CString(self, "previous_disk", "Description of disk/media on which previous cabinet resides", charset="ASCII")
        if self["flags/has_next"].value:
            yield CString(self, "next_cabinet", "File name of next cabinet", charset="ASCII")
            yield CString(self, "next_disk", "Description of disk/media on which next cabinet resides", charset="ASCII")

        folders = []
        files = []
        for index in range(self["nb_folder"].value):
            folder = Folder(self, "folder[]")
            yield folder
            folders.append(folder)
        for index in range(self["nb_files"].value):
            file = File(self, "file[]")
            yield file
            files.append(file)

        folders = sorted(enumerate(folders),
                         key=lambda x: x[1]["offset"].value)

        for i in range(len(folders)):
            index, folder = folders[i]
            padding = self.seekByte(folder["offset"].value)
            if padding:
                yield padding
            files = []
            for file in files:
                if file["folder_index"].value == index:
                    files.append(file)
            if i + 1 == len(folders):
                size = (self.size // 8) - folder["offset"].value
            else:
                size = (folders[i + 1][1]["offset"].value) - \
                    folder["offset"].value
            yield FolderData(self, "folder_data[%i]" % index, folder, files, size=size * 8)

        end = self.seekBit(self.size, "endraw")
        if end:
            yield end

    def createContentSize(self):
        return self["filesize"].value * 8



"""
Apple/NeXT Binary Property List (BPLIST) parser.

Also includes a .createXML() function which produces an XML representation of the object.
Note that it will discard unknown objects, nulls and fill values, but should work for most files.

Documents:
- CFBinaryPList.c
  http://src.gnu-darwin.org/DarwinSourceArchive/expanded/CF/CF-299/Parsing.subproj/CFBinaryPList.c
- ForFoundationOnly.h (for structure formats)
  http://src.gnu-darwin.org/DarwinSourceArchive/expanded/CF/CF-299/Base.subproj/ForFoundationOnly.h
- XML <-> BPList converter
  http://scw.us/iPhone/plutil/plutil.pl
Author: Robert Xiao
Created: 2008-09-21
"""

from hachoir.parser import HachoirParser
from hachoir.field import (RootSeekableFieldSet, FieldSet, Enum,
                           Bits, GenericInteger, Float32, Float64, UInt8, UInt64, Bytes, NullBytes, RawBytes, String)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import displayHandler
from hachoir.core.tools import humanDatetime
from datetime import datetime, timedelta


class BPListTrailer(FieldSet):

    def createFields(self):
        yield NullBytes(self, "unused", 6)
        yield UInt8(self, "offsetIntSize", "Size (in bytes) of offsets in the offset table")
        yield UInt8(self, "objectRefSize", "Size (in bytes) of object numbers in object references")
        yield UInt64(self, "numObjects", "Number of objects in this file")
        yield UInt64(self, "topObject", "Top-level object reference")
        yield UInt64(self, "offsetTableOffset", "File offset to the offset table")

    def createDescription(self):
        return "Binary PList trailer"


class BPListOffsetTable(FieldSet):

    def createFields(self):
        size = self["../trailer/offsetIntSize"].value * 8
        for i in range(self["../trailer/numObjects"].value):
            yield Bits(self, "offset[]", size)


class BPListSize(FieldSet):

    def createFields(self):
        yield Bits(self, "size", 4)
        if self['size'].value == 0xF:
            yield BPListObject(self, "fullsize")

    def createValue(self):
        if 'fullsize' in self:
            return self['fullsize'].value
        else:
            return self['size'].value


class BPListObjectRef(GenericInteger):

    def __init__(self, parent, name, description=None):
        size = parent['/trailer/objectRefSize'].value * 8
        GenericInteger.__init__(self, parent, name, False, size, description)

    def getRef(self):
        return self.parent['/object[' + str(self.value) + ']']

    def createDisplay(self):
        return self.getRef().display

    def createXML(self, prefix=''):
        return self.getRef().createXML(prefix)


class BPListArray(FieldSet):

    def __init__(self, parent, name, size, description=None):
        FieldSet.__init__(self, parent, name, description=description)
        self.numels = size

    def createFields(self):
        for i in range(self.numels):
            yield BPListObjectRef(self, "ref[]")

    def createValue(self):
        return self.array('ref')

    def createDisplay(self):
        return '[' + ', '.join([x.display for x in self.value]) + ']'

    def createXML(self, prefix=''):
        return prefix + '<array>\n' + ''.join([x.createXML(prefix + '\t') + '\n' for x in self.value]) + prefix + '</array>'


class BPListDict(FieldSet):

    def __init__(self, parent, name, size, description=None):
        FieldSet.__init__(self, parent, name, description=description)
        self.numels = size

    def createFields(self):
        for i in range(self.numels):
            yield BPListObjectRef(self, "keyref[]")
        for i in range(self.numels):
            yield BPListObjectRef(self, "valref[]")

    def createValue(self):
        return list(zip(self.array('keyref'), self.array('valref')))

    def createDisplay(self):
        return '{' + ', '.join(['%s: %s' % (k.display, v.display) for k, v in self.value]) + '}'

    def createXML(self, prefix=''):
        return prefix + '<dict>\n' + ''.join(['%s\t<key>%s</key>\n%s\n' % (prefix, k.getRef().value.encode('utf-8'), v.createXML(prefix + '\t')) for k, v in self.value]) + prefix + '</dict>'


class BPListObject(FieldSet):

    def createFields(self):
        yield Enum(Bits(self, "marker_type", 4),
                   {0: "Simple",
                    1: "Int",
                    2: "Real",
                    3: "Date",
                    4: "Data",
                    5: "ASCII String",
                    6: "UTF-16-BE String",
                    8: "UID",
                    10: "Array",
                    13: "Dict", })
        markertype = self['marker_type'].value
        if markertype == 0:
            # Simple (Null)
            yield Enum(Bits(self, "value", 4),
                       {0: "Null",
                        8: "False",
                        9: "True",
                        15: "Fill Byte", })
            if self['value'].display == "False":
                self.xml = lambda prefix: prefix + "<false/>"
            elif self['value'].display == "True":
                self.xml = lambda prefix: prefix + "<true/>"
            else:
                self.xml = lambda prefix: prefix + ""

        elif markertype == 1:
            # Int
            yield Bits(self, "size", 4, "log2 of number of bytes")
            size = self['size'].value
            # 8-bit (size=0), 16-bit (size=1) and 32-bit (size=2) numbers are unsigned
            # 64-bit (size=3) numbers are signed
            yield GenericInteger(self, "value", (size >= 3), (2**size) * 8)
            self.xml = lambda prefix: prefix + \
                "<integer>%s</integer>" % self['value'].value

        elif markertype == 2:
            # Real
            yield Bits(self, "size", 4, "log2 of number of bytes")
            if self['size'].value == 2:  # 2**2 = 4 byte float
                yield Float32(self, "value")
            elif self['size'].value == 3:  # 2**3 = 8 byte float
                yield Float64(self, "value")
            else:
                # FIXME: What is the format of the real?
                yield Bits(self, "value", (2**self['size'].value) * 8)
            self.xml = lambda prefix: prefix + \
                "<real>%s</real>" % self['value'].value

        elif markertype == 3:
            # Date
            yield Bits(self, "extra", 4, "Extra value, should be 3")
            # Use a heuristic to determine which epoch to use

            def cvt_time(v):
                v = timedelta(seconds=v)
                epoch2001 = datetime(2001, 1, 1)
                epoch1970 = datetime(1970, 1, 1)
                if (epoch2001 + v - datetime.today()).days > 5 * 365:
                    return epoch1970 + v
                return epoch2001 + v
            yield displayHandler(Float64(self, "value"), lambda x: humanDatetime(cvt_time(x)))
            self.xml = lambda prefix: prefix + \
                "<date>%sZ</date>" % (
                    cvt_time(self['value'].value).isoformat())

        elif markertype == 4:
            # Data
            yield BPListSize(self, "size")
            if self['size'].value:
                yield Bytes(self, "value", self['size'].value)
                self.xml = lambda prefix: prefix + \
                    "<data>\n%s\n%s</data>" % (
                        self['value'].value.encode('base64').strip(), prefix)
            else:
                self.xml = lambda prefix: prefix + '<data></data>'

        elif markertype == 5:
            # ASCII String
            yield BPListSize(self, "size")
            if self['size'].value:
                yield String(self, "value", self['size'].value, charset="ASCII")
                self.xml = lambda prefix: prefix + \
                    "<string>%s</string>" % (self['value'].value.replace(
                        '&', '&amp;').encode('iso-8859-1'))
            else:
                self.xml = lambda prefix: prefix + '<string></string>'

        elif markertype == 6:
            # UTF-16-BE String
            yield BPListSize(self, "size")
            if self['size'].value:
                yield String(self, "value", self['size'].value * 2, charset="UTF-16-BE")
                self.xml = lambda prefix: prefix + \
                    "<string>%s</string>" % (
                        self['value'].value.replace('&', '&amp;').encode('utf-8'))
            else:
                self.xml = lambda prefix: prefix + '<string></string>'

        elif markertype == 8:
            # UID
            yield Bits(self, "size", 4, "Number of bytes minus 1")
            yield GenericInteger(self, "value", False, (self['size'].value + 1) * 8)
            self.xml = lambda prefix: prefix + ""  # no equivalent?

        elif markertype == 10:
            # Array
            yield BPListSize(self, "size")
            size = self['size'].value
            if size:
                yield BPListArray(self, "value", size)
                self.xml = lambda prefix: self['value'].createXML(prefix)

        elif markertype == 13:
            # Dict
            yield BPListSize(self, "size")
            yield BPListDict(self, "value", self['size'].value)
            self.xml = lambda prefix: self['value'].createXML(prefix)

        else:
            yield Bits(self, "value", 4)
            self.xml = lambda prefix: ''

    def createValue(self):
        if 'value' in self:
            return self['value'].value
        elif self['marker_type'].value in [4, 5, 6]:
            return ''
        else:
            return None

    def createDisplay(self):
        if 'value' in self:
            return str(self['value'].display)
        elif self['marker_type'].value in [4, 5, 6]:
            return ''
        else:
            return None

    def createXML(self, prefix=''):
        if 'value' in self:
            try:
                return self.xml(prefix)
            except AttributeError:
                return ''
        return ''

    def getFieldType(self):
        return '%s<%s>' % (FieldSet.getFieldType(self), self['marker_type'].display)


class BPList(HachoirParser, RootSeekableFieldSet):
    endian = BIG_ENDIAN
    MAGIC = b"bplist00"
    PARSER_TAGS = {
        "id": "bplist",
        "category": "misc",
        "file_ext": ("plist",),
        "magic": ((MAGIC, 0),),
        "min_size": 8 + 32,  # bplist00 + 32-byte trailer
        "description": "Apple/NeXT Binary Property List",
    }

    def __init__(self, stream, **args):
        RootSeekableFieldSet.__init__(
            self, None, "root", stream, None, stream.askSize(self))
        HachoirParser.__init__(self, stream, **args)

    def validate(self):
        if self.stream.readBytes(0, len(self.MAGIC)) != self.MAGIC:
            return "Invalid magic"
        return True

    def createFields(self):
        yield Bytes(self, "magic", 8, "File magic (bplist00)")
        if self.size:
            self.seekByte(self.size // 8 - 32, True)
        else:
            # FIXME: UNTESTED
            while True:
                try:
                    self.seekByte(1024)
                except Exception:
                    break
            self.seekByte(self.size // 8 - 32)
        yield BPListTrailer(self, "trailer")
        self.seekByte(self['trailer/offsetTableOffset'].value)
        yield BPListOffsetTable(self, "offset_table")
        for i in self.array("offset_table/offset"):
            if self.current_size > i.value * 8:
                self.seekByte(i.value)
            elif self.current_size < i.value * 8:
                # try to detect files with gaps or unparsed content
                yield RawBytes(self, "padding[]", i.value - self.current_size // 8)
            yield BPListObject(self, "object[]")

    def createXML(self, prefix=''):
        return '''<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple Computer//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
''' + self['/object[' + str(self['/trailer/topObject'].value) + ']'].createXML(prefix) + '''
</plist>'''



from hotshot import Profile
from hotshot.stats import load as loadStats
from os import unlink


def runProfiler(func, args=tuple(), kw={}, verbose=True, nb_func=25,
                sort_by=('cumulative', 'calls')):
    profile_filename = "/tmp/profiler"
    prof = Profile(profile_filename)
    try:
        if verbose:
            print("[+] Run profiler")
        result = prof.runcall(func, *args, **kw)
        prof.close()
        if verbose:
            print("[+] Stop profiler")
            print("[+] Process data...")
        stat = loadStats(profile_filename)
        if verbose:
            print("[+] Strip...")
        stat.strip_dirs()
        if verbose:
            print("[+] Sort data...")
        stat.sort_stats(*sort_by)
        if verbose:
            print()
            print("[+] Display statistics")
            print()
        stat.print_stats(nb_func)
        return result
    finally:
        unlink(profile_filename)



"""
Microsoft Bitmap picture parser.
- file extension: ".bmp"

Author: Victor Stinner
Creation: 16 december 2005
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt8, UInt16, UInt32, Bits,
                           String, RawBytes, Enum,
                           PaddingBytes, NullBytes, createPaddingField)
from hachoir.core.endian import LITTLE_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.parser.image.common import RGB, PaletteRGBA
from hachoir.core.tools import alignValue


class Pixel4bit(Bits):
    static_size = 4

    def __init__(self, parent, name):
        Bits.__init__(self, parent, name, 4)


class ImageLine(FieldSet):

    def __init__(self, parent, name, width, pixel_class):
        FieldSet.__init__(self, parent, name)
        self._pixel = pixel_class
        self._width = width
        self._size = alignValue(self._width * self._pixel.static_size, 32)

    def createFields(self):
        for x in range(self._width):
            yield self._pixel(self, "pixel[]")
        size = self.size - self.current_size
        if size:
            yield createPaddingField(self, size)


class ImagePixels(FieldSet):

    def __init__(self, parent, name, width, height, pixel_class, size=None):
        FieldSet.__init__(self, parent, name, size=size)
        self._width = width
        self._height = height
        self._pixel = pixel_class

    def createFields(self):
        for y in range(self._height - 1, -1, -1):
            yield ImageLine(self, "line[%u]" % y, self._width, self._pixel)
        size = (self.size - self.current_size) // 8
        if size:
            yield NullBytes(self, "padding", size)


class CIEXYZ(FieldSet):

    def createFields(self):
        yield UInt32(self, "x")
        yield UInt32(self, "y")
        yield UInt32(self, "z")


class BmpHeader(FieldSet):
    color_space_name = {
        1: "Business (Saturation)",
        2: "Graphics (Relative)",
        4: "Images (Perceptual)",
        8: "Absolute colormetric (Absolute)",
    }

    def getFormatVersion(self):
        if "gamma_blue" in self:
            return 4
        if "important_color" in self:
            return 3
        return 2

    def createFields(self):
        # Version 2 (12 bytes)
        yield UInt32(self, "header_size", "Header size")
        yield UInt32(self, "width", "Width (pixels)")
        yield UInt32(self, "height", "Height (pixels)")
        yield UInt16(self, "nb_plan", "Number of plan (=1)")
        # may be zero for PNG/JPEG picture
        yield UInt16(self, "bpp", "Bits per pixel")

        # Version 3 (40 bytes)
        if self["header_size"].value < 40:
            return
        yield Enum(UInt32(self, "compression", "Compression method"), BmpFile.COMPRESSION_NAME)
        yield UInt32(self, "image_size", "Image size (bytes)")
        yield UInt32(self, "horizontal_dpi", "Horizontal DPI")
        yield UInt32(self, "vertical_dpi", "Vertical DPI")
        yield UInt32(self, "used_colors", "Number of color used")
        yield UInt32(self, "important_color", "Number of import colors")

        # Version 4 (108 bytes)
        if self["header_size"].value < 108:
            return
        yield textHandler(UInt32(self, "red_mask"), hexadecimal)
        yield textHandler(UInt32(self, "green_mask"), hexadecimal)
        yield textHandler(UInt32(self, "blue_mask"), hexadecimal)
        yield textHandler(UInt32(self, "alpha_mask"), hexadecimal)
        yield Enum(UInt32(self, "color_space"), self.color_space_name)
        yield CIEXYZ(self, "red_primary")
        yield CIEXYZ(self, "green_primary")
        yield CIEXYZ(self, "blue_primary")
        yield UInt32(self, "gamma_red")
        yield UInt32(self, "gamma_green")
        yield UInt32(self, "gamma_blue")


def parseImageData(parent, name, size, header):
    if ("compression" not in header) or (header["compression"].value in (0, 3)):
        width = header["width"].value
        height = header["height"].value
        bpp = header["bpp"].value
        if bpp == 32:
            cls = UInt32
        elif bpp == 24:
            cls = RGB
        elif bpp == 8:
            cls = UInt8
        elif bpp == 4:
            cls = Pixel4bit
        else:
            cls = None
        if cls:
            return ImagePixels(parent, name, width, height, cls, size=size * 8)
    return RawBytes(parent, name, size)


class BmpFile(Parser):
    PARSER_TAGS = {
        "id": "bmp",
        "category": "image",
        "file_ext": ("bmp",),
        "mime": ("image/x-ms-bmp", "image/x-bmp"),
        "min_size": 30 * 8,
        #        "magic": ((b"BM", 0),),
        "magic_regex": ((
            # "BM", <filesize>, <reserved>, header_size=(12|40|108)
            b"BM.{4}.{8}[\x0C\x28\x6C]\0{3}",
            0),),
        "description": "Microsoft bitmap (BMP) picture"
    }
    endian = LITTLE_ENDIAN

    COMPRESSION_NAME = {
        0: "Uncompressed",
        1: "RLE 8-bit",
        2: "RLE 4-bit",
        3: "Bitfields",
        4: "JPEG",
        5: "PNG",
    }

    def validate(self):
        if self.stream.readBytes(0, 2) != b'BM':
            return "Wrong file signature"
        if self["header/header_size"].value not in (12, 40, 108):
            return "Unknown header size (%s)" % self["header_size"].value
        if self["header/nb_plan"].value != 1:
            return "Invalid number of planes"
        return True

    def createFields(self):
        yield String(self, "signature", 2, "Header (\"BM\")", charset="ASCII")
        yield UInt32(self, "file_size", "File size (bytes)")
        yield PaddingBytes(self, "reserved", 4, "Reserved")
        yield UInt32(self, "data_start", "Data start position")
        yield BmpHeader(self, "header")

        # Compute number of color
        header = self["header"]
        bpp = header["bpp"].value
        if 0 < bpp <= 8:
            if "used_colors" in header and header["used_colors"].value:
                nb_color = header["used_colors"].value
            else:
                nb_color = (1 << bpp)
        else:
            nb_color = 0

        # Color palette (if any)
        if nb_color:
            yield PaletteRGBA(self, "palette", nb_color)

        # Seek to data start
        field = self.seekByte(self["data_start"].value)
        if field:
            yield field

        # Image pixels
        size = min(self["file_size"].value - self["data_start"].value,
                   (self.size - self.current_size) // 8)
        yield parseImageData(self, "pixels", size, header)

    def createDescription(self):
        return "Microsoft Bitmap version %s" % self["header"].getFormatVersion()

    def createContentSize(self):
        return self["file_size"].value * 8



from hachoir.metadata.metadata import RootMetadata, registerExtractor
from hachoir.metadata.safe import fault_tolerant
from hachoir.parser.file_system import ISO9660
from datetime import datetime


class ISO9660_Metadata(RootMetadata):

    def extract(self, iso):
        desc = iso['volume[0]/content']
        self.title = desc['volume_id'].value
        self.title = desc['vol_set_id'].value
        self.author = desc['publisher'].value
        self.author = desc['data_preparer'].value
        self.producer = desc['application'].value
        self.copyright = desc['copyright'].value
        self.readTimestamp('creation_date', desc['creation_ts'].value)
        self.readTimestamp('last_modification', desc['modification_ts'].value)

    @fault_tolerant
    def readTimestamp(self, key, value):
        if value.startswith("0000"):
            return
        value = datetime(
            int(value[0:4]), int(value[4:6]), int(value[6:8]),
            int(value[8:10]), int(value[10:12]), int(value[12:14]))
        setattr(self, key, value)


registerExtractor(ISO9660, ISO9660_Metadata)



"""
Musical Instrument Digital Interface (MIDI) audio file parser.

Documentation:
 - Standard MIDI File Format, Dustin Caldwell (downloaded on wotsit.org)

Author: Victor Stinner
Creation: 27 december 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, Bits, ParserError,
                           String, UInt32, UInt24, UInt16, UInt8, Enum, RawBits, RawBytes)
from hachoir.core.endian import BIG_ENDIAN
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.core.tools import createDict, humanDurationNanosec
from hachoir.parser.common.tracker import NOTE_NAME

MAX_FILESIZE = 10 * 1024 * 1024


class Integer(Bits):

    def __init__(self, parent, name, description=None):
        Bits.__init__(self, parent, name, 8, description)
        stream = parent.stream
        addr = self.absolute_address
        value = 0
        while True:
            bits = stream.readBits(addr, 8, parent.endian)
            value = (value << 7) + (bits & 127)
            if not (bits & 128):
                break
            addr += 8
            self._size += 8
            if 32 < self._size:
                raise ParserError("Integer size is bigger than 32-bit")
        self.createValue = lambda: value


def parseNote(parser):
    yield Enum(UInt8(parser, "note", "Note number"), NOTE_NAME)
    yield UInt8(parser, "velocity")


def parseControl(parser):
    yield UInt8(parser, "control", "Controller number")
    yield UInt8(parser, "value", "New value")


def parsePatch(parser):
    yield UInt8(parser, "program", "New program number")


def parseChannel(parser, size=1):
    yield UInt8(parser, "channel", "Channel number")


def parsePitch(parser):
    yield UInt8(parser, "bottom", "(least sig) 7 bits of value")
    yield UInt8(parser, "top", "(most sig) 7 bits of value")


def parseText(parser, size):
    yield String(parser, "text", size)


def parseSMPTEOffset(parser, size):
    yield RawBits(parser, "padding", 1)
    yield Enum(Bits(parser, "frame_rate", 2),
               {0: "24 fps", 1: "25 fps", 2: "30 fps (drop frame)", 3: "30 fps"})
    yield Bits(parser, "hour", 5)
    yield UInt8(parser, "minute")
    yield UInt8(parser, "second")
    yield UInt8(parser, "frame")
    yield UInt8(parser, "subframe", "100 subframes per frame")


def formatTempo(field):
    return humanDurationNanosec(field.value * 1000)


def parseTempo(parser, size):
    yield textHandler(UInt24(parser, "microsec_quarter", "Microseconds per quarter note"), formatTempo)


def parseTimeSignature(parser, size):
    yield UInt8(parser, "numerator", "Numerator of time signature")
    yield UInt8(parser, "denominator", "denominator of time signature 2=quarter 3=eighth, etc.")
    yield UInt8(parser, "nb_tick", "Number of ticks in metronome click")
    yield UInt8(parser, "nb_32nd_note", "Number of 32nd notes to the quarter note")


class Command(FieldSet):
    COMMAND = {}
    for channel in range(16):
        COMMAND[0x80 + channel] = ("Note off (channel %u)" %
                                   channel, parseNote)
        COMMAND[0x90 + channel] = ("Note on (channel %u)" % channel, parseNote)
        COMMAND[
            0xA0 + channel] = ("Key after-touch (channel %u)" % channel, parseNote)
        COMMAND[0xB0 + channel] = ("Control change (channel %u)" %
                                   channel, parseControl)
        COMMAND[
            0xC0 + channel] = ("Program (patch) change (channel %u)" % channel, parsePatch)
        COMMAND[
            0xD0 + channel] = ("Channel after-touch (channel %u)" % channel, parseChannel)
        COMMAND[
            0xE0 + channel] = ("Pitch wheel change (channel %u)" % channel, parsePitch)
    COMMAND_DESC = createDict(COMMAND, 0)
    COMMAND_PARSER = createDict(COMMAND, 1)

    META_COMMAND_TEXT = 1
    META_COMMAND_NAME = 3
    META_COMMAND = {
        0x00: ("Sets the track's sequence number", None),
        0x01: ("Text event", parseText),
        0x02: ("Copyright info", parseText),
        0x03: ("Sequence or Track name", parseText),
        0x04: ("Track instrument name", parseText),
        0x05: ("Lyric", parseText),
        0x06: ("Marker", parseText),
        0x07: ("Cue point", parseText),
        0x20: ("MIDI Channel Prefix", parseChannel),
        0x2F: ("End of the track", None),
        0x51: ("Set tempo", parseTempo),
        0x54: ("SMPTE offset", parseSMPTEOffset),
        0x58: ("Time Signature", parseTimeSignature),
        0x59: ("Key signature", None),
        0x7F: ("Sequencer specific information", None),
    }
    META_COMMAND_DESC = createDict(META_COMMAND, 0)
    META_COMMAND_PARSER = createDict(META_COMMAND, 1)

    def __init__(self, *args, **kwargs):
        if 'prev_command' in kwargs:
            self.prev_command = kwargs['prev_command']
            del kwargs['prev_command']
        else:
            self.prev_command = None
        self.command = None
        FieldSet.__init__(self, *args, **kwargs)

    def createFields(self):
        yield Integer(self, "time", "Delta time in ticks")
        next = self.stream.readBits(
            self.absolute_address + self.current_size, 8, self.root.endian)
        if next & 0x80 == 0:
            # "Running Status" command
            if self.prev_command is None:
                raise ParserError(
                    "Running Status command not preceded by another command.")
            self.command = self.prev_command.command
        else:
            yield Enum(textHandler(UInt8(self, "command"), hexadecimal), self.COMMAND_DESC)
            self.command = self["command"].value
        if self.command == 0xFF:
            yield Enum(textHandler(UInt8(self, "meta_command"), hexadecimal), self.META_COMMAND_DESC)
            yield UInt8(self, "data_len")
            size = self["data_len"].value
            if size:
                command = self["meta_command"].value
                if command in self.META_COMMAND_PARSER:
                    parser = self.META_COMMAND_PARSER[command]
                else:
                    parser = None
                if parser is not None:
                    yield from parser(self, size)
                else:
                    yield RawBytes(self, "data", size)
        else:
            if self.command not in self.COMMAND_PARSER:
                raise ParserError("Unknown command: %s" %
                                  self["command"].display)
            parser = self.COMMAND_PARSER[self.command]
            yield from parser(self)

    def createDescription(self):
        if "meta_command" in self:
            return self["meta_command"].display
        else:
            return self.COMMAND_DESC[self.command]


class Track(FieldSet):

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        self._size = (8 + self["size"].value) * 8

    def createFields(self):
        yield String(self, "marker", 4, "Track marker (MTrk)", charset="ASCII")
        yield UInt32(self, "size")
        cur = None
        if True:
            while not self.eof:
                cur = Command(self, "command[]", prev_command=cur)
                yield cur
        else:
            size = self["size"].value
            if size:
                yield RawBytes(self, "raw", size)

    def createDescription(self):
        command = self["command[0]"]
        if "meta_command" in command \
                and command["meta_command"].value in (Command.META_COMMAND_TEXT, Command.META_COMMAND_NAME) \
                and "text" in command:
            return command["text"].value.strip("\r\n")
        else:
            return ""


class Header(FieldSet):
    static_size = 10 * 8
    FILE_FORMAT = {
        0: "Single track",
        1: "Multiple tracks, synchronous",
        2: "Multiple tracks, asynchronous",
    }

    def createFields(self):
        yield UInt32(self, "size")
        yield Enum(UInt16(self, "file_format"), self.FILE_FORMAT)
        yield UInt16(self, "nb_track")
        yield UInt16(self, "delta_time", "Delta-time ticks per quarter note")

    def createDescription(self):
        return "%s; %s tracks" % (
            self["file_format"].display, self["nb_track"].value)


class MidiFile(Parser):
    MAGIC = b"MThd"
    PARSER_TAGS = {
        "id": "midi",
        "category": "audio",
        "file_ext": ["mid", "midi"],
        "mime": ("audio/mime", ),
        "magic": ((MAGIC, 0),),
        "min_size": 64,
        "description": "MIDI audio"
    }
    endian = BIG_ENDIAN

    def validate(self):
        if self.stream.readBytes(0, 4) != self.MAGIC:
            return "Invalid signature"
        if self["header/size"].value != 6:
            return "Invalid header size"
        return True

    def createFields(self):
        yield String(self, "signature", 4, r"MIDI signature (MThd)", charset="ASCII")
        yield Header(self, "header")
        while not self.eof:
            yield Track(self, "track[]")

    def createDescription(self):
        return "MIDI audio: %s" % self["header"].description

    def createContentSize(self):
        count = self["/header/nb_track"].value - 1
        start = self["track[%u]" % count].absolute_address
        # Search "End of track" of last track
        end = self.stream.searchBytes(b"\xff\x2f\x00", start, MAX_FILESIZE * 8)
        if end is not None:
            return end + 3 * 8
        return None



#!/usr/bin/env python3

from hachoir.wx.app import app_t
from hachoir import __version__
from hachoir.core.cmd_line import getHachoirOptions, configureHachoir
from optparse import OptionParser
import sys


def parseOptions():
    parser = OptionParser(usage="%prog [options] [filename]")
    hachoir = getHachoirOptions(parser)
    parser.add_option_group(hachoir)

    values, arguments = parser.parse_args()
    if len(arguments) == 1:
        filename = arguments[0]
    elif not arguments:
        filename = None
    else:
        parser.print_help()
        sys.exit(1)
    return values, filename


def main():
    print("hachoir version %s" % __version__)
    print()
    values, filename = parseOptions()
    configureHachoir(values)
    app = app_t(filename)
    app.MainLoop()


if __name__ == '__main__':
    main()



"""
Integer field classes:
- UInt8, UInt16, UInt24, UInt32, UInt64: unsigned integer of 8, 16, 32, 64 bits ;
- Int8, Int16, Int24, Int32, Int64: signed integer of 8, 16, 32, 64 bits.
"""

from hachoir.field import Bits, FieldError


class GenericInteger(Bits):
    """
    Generic integer class used to generate other classes.
    """

    def __init__(self, parent, name, signed, size, description=None):
        if not (8 <= size <= 16384):
            raise FieldError(
                "Invalid integer size (%s): have to be in 8..16384" % size)
        Bits.__init__(self, parent, name, size, description)
        self.signed = signed

    def createValue(self):
        return self._parent.stream.readInteger(
            self.absolute_address, self.signed, self._size, self._parent.endian)


def integerFactory(name, is_signed, size, doc):
    class Integer(GenericInteger):
        __doc__ = doc
        static_size = size

        def __init__(self, parent, name, description=None):
            GenericInteger.__init__(
                self, parent, name, is_signed, size, description)
    cls = Integer
    cls.__name__ = name
    return cls


UInt8 = integerFactory("UInt8", False, 8, "Unsigned integer of 8 bits")
UInt16 = integerFactory("UInt16", False, 16, "Unsigned integer of 16 bits")
UInt24 = integerFactory("UInt24", False, 24, "Unsigned integer of 24 bits")
UInt32 = integerFactory("UInt32", False, 32, "Unsigned integer of 32 bits")
UInt64 = integerFactory("UInt64", False, 64, "Unsigned integer of 64 bits")

Int8 = integerFactory("Int8", True, 8, "Signed integer of 8 bits")
Int16 = integerFactory("Int16", True, 16, "Signed integer of 16 bits")
Int24 = integerFactory("Int24", True, 24, "Signed integer of 24 bits")
Int32 = integerFactory("Int32", True, 32, "Signed integer of 32 bits")
Int64 = integerFactory("Int64", True, 64, "Signed integer of 64 bits")



import wx


class field_split_menu_fwd_t:
    def __init__(self, imp):
        self.imp = imp

    def on_field_menu_ready(self, dispatcher, view):
        assert view is not None

        view.Bind(wx.EVT_MENU, self.on_split_bytes,
                  id=wx.xrc.XRCID('field_menu_split_bytes'))
        view.Bind(wx.EVT_MENU, self.on_split_bits,
                  id=wx.xrc.XRCID('field_menu_split_bits'))

    def on_split_bits(self, event):
        self.imp.on_split_bits()

    def on_split_bytes(self, event):
        self.imp.on_split_bytes()



"""
Parsers for the different streams and fragments found in an OLE2 file.

Documents:
 - goffice source code
 - Microsoft Office PowerPoint 97-2007 Binary File Format (.ppt) Specification
    http://download.microsoft.com/download/0/B/E/0BE8BDD7-E5E8-422A-ABFD-4342ED7AD886/PowerPoint97-2007BinaryFileFormat(ppt)Specification.pdf

Author: Robert Xiao, Victor Stinner
Creation: 8 january 2005
"""

from hachoir.field import (SubFile, FieldSet,
                           UInt8, UInt16, UInt32, Enum, String, CString,
                           Bits, RawBytes, CustomFragment)
from hachoir.core.text_handler import textHandler, hexadecimal
from hachoir.parser.misc.ole2_util import OLE2FragmentParser, RawParser
from hachoir.parser.misc.msoffice_summary import Summary, DocSummary, CompObj
from hachoir.parser.misc.word_doc import WordDocumentParser, WordTableParser


class RootEntry(OLE2FragmentParser):
    ENDIAN_CHECK = False

    def createFields(self):
        for index, property in enumerate(self.ole2.properties):
            if index == 0:
                continue
            try:
                name, parser = PROPERTY_NAME[property["name"].value]
            except LookupError:
                name = property.name + "content"
                parser = RawParser
            yield from self.parseProperty(property, name, parser)

    def seekSBlock(self, block):
        self.seekBit(block * self.ole2.ss_size)

    def parseProperty(self, property, name_prefix, parser=RawParser):
        ole2 = self.ole2
        if not property["size"].value:
            return
        if property["size"].value >= ole2["header/threshold"].value:
            return
        name = "%s[]" % name_prefix
        first = None
        previous = None
        size = 0
        fragment_group = None
        chain = ole2.getChain(property["start"].value, ole2.ss_fat)
        while True:
            try:
                block = next(chain)
                contiguous = False
                if first is None:
                    first = block
                    contiguous = True
                if previous is not None and block == (previous + 1):
                    contiguous = True
                if contiguous:
                    previous = block
                    size += ole2.ss_size
                    continue
            except StopIteration:
                block = None
            if first is None:
                break
            self.seekSBlock(first)
            desc = "Small blocks %s..%s (%s)" % (
                first, previous, previous - first + 1)
            desc += " of %s bytes" % (ole2.ss_size // 8)
            field = CustomFragment(
                self, name, size, parser, desc, fragment_group)
            yield field
            if not fragment_group:
                fragment_group = field.group
                fragment_group.args["ole2"] = field.root
                fragment_group.args["datasize"] = property["size"].value
                fragment_group.args["ole2name"] = property["name"].value
            if block is None:
                break
            first = block
            previous = block
            size = ole2.ss_size


class Pictures(OLE2FragmentParser):

    class Picture(FieldSet):

        def createFields(self):
            yield RawBytes(self, "identifier", 4, "some kind of marker (A0461DF0)")
            yield UInt32(self, "size")
            yield RawBytes(self, "unknown[]", 16)
            yield RawBytes(self, "unknown[]", 1)
            yield SubFile(self, "image", self["size"].value - 17, "Image Data")
    ENDIAN_CHECK = False

    def createFields(self):
        pos = 0
        while pos // 8 < self.datasize:
            newpic = Pictures.Picture(self, "picture[]")
            yield newpic
            pos += newpic.size


class PowerPointDocument(OLE2FragmentParser):
    OBJ_TYPES = {0: "Unknown",
                 1000: "Document",
                 1001: "DocumentAtom",
                 1002: "EndDocument",
                 1003: "SlidePersist",
                 1004: "SlideBase",
                 1005: "SlideBaseAtom",
                 1006: "Slide",
                 1007: "SlideAtom",
                 1008: "Notes",
                 1009: "NotesAtom",
                 1010: "Environment",
                 1011: "SlidePersistAtom",
                 1012: "Scheme",
                 1013: "SchemeAtom",
                 1014: "DocViewInfo",
                 1015: "SSlideLayoutAtom",
                 1016: "MainMaster",
                 1017: "SSSlideInfoAtom",
                 1018: "SlideViewInfo",
                 1019: "GuideAtom",
                 1020: "ViewInfo",
                 1021: "ViewInfoAtom",
                 1022: "SlideViewInfoAtom",
                 1023: "VBAInfo",
                 1024: "VBAInfoAtom",
                 1025: "SSDocInfoAtom",
                 1026: "Summary",
                 1027: "Texture",
                 1028: "VBASlideInfo",
                 1029: "VBASlideInfoAtom",
                 1030: "DocRoutingSlip",
                 1031: "OutlineViewInfo",
                 1032: "SorterViewInfo",
                 1033: "ExObjList",
                 1034: "ExObjListAtom",
                 1035: "PPDrawingGroup",  # FIXME: Office Art File Format Docu
                 1036: "PPDrawing",  # FIXME: Office Art File Format Docu
                 1038: "Theme",
                 1039: "ColorMapping",
                 1040: "NamedShows",  # don't know if container
                 1041: "NamedShow",
                 1042: "NamedShowSlides",  # don't know if container
                 1052: "OriginalMainMasterId",
                 1053: "CompositeMasterId",
                 1054: "RoundTripContentMasterInfo12",
                 1055: "RoundTripShapeId12",
                 1056: "RoundTripHFPlaceholder12",
                 1058: "RoundTripContentMasterId12",
                 1059: "RoundTripOArtTextStyles12",
                 1060: "HeaderFooterDefaults12",
                 1061: "DocFlags12",
                 1062: "RoundTripShapeCheckSumForCustomLayouts12",
                 1063: "RoundTripNotesMasterTextStyles12",
                 1064: "RoundTripCustomTableStyles12",
                 2000: "List",
                 2005: "FontCollection",
                 2017: "ListPlaceholder",
                 2019: "BookmarkCollection",
                 2020: "SoundCollection",
                 2021: "SoundCollAtom",
                 2022: "Sound",
                 2023: "SoundData",
                 2025: "BookmarkSeedAtom",
                 2026: "GuideList",
                 2028: "RunArray",
                 2029: "RunArrayAtom",
                 2030: "ArrayElementAtom",
                 2031: "Int4ArrayAtom",
                 2032: "ColorSchemeAtom",
                 3008: "OEShape",
                 3009: "ExObjRefAtom",
                 3011: "OEPlaceholderAtom",
                 3020: "GrColor",
                 3024: "GPointAtom",
                 3025: "GrectAtom",
                 3031: "GRatioAtom",
                 3032: "Gscaling",
                 3034: "GpointAtom",
                 3035: "OEShapeAtom",
                 3037: "OEPlaceholderNewPlaceholderId12",
                 3998: "OutlineTextRefAtom",
                 3999: "TextHeaderAtom",
                 4000: "TextCharsAtom",
                 4001: "StyleTextPropAtom",
                 4002: "BaseTextPropAtom",
                 4003: "TxMasterStyleAtom",
                 4004: "TxCFStyleAtom",
                 4005: "TxPFStyleAtom",
                 4006: "TextRulerAtom",
                 4007: "TextBookmarkAtom",
                 4008: "TextBytesAtom",
                 4009: "TxSIStyleAtom",
                 4010: "TextSpecInfoAtom",
                 4011: "DefaultRulerAtom",
                 4023: "FontEntityAtom",
                 4024: "FontEmbeddedData",
                 4025: "TypeFace",
                 4026: "CString",
                 4027: "ExternalObject",
                 4033: "MetaFile",
                 4034: "ExOleObj",
                 4035: "ExOleObjAtom",
                 4036: "ExPlainLinkAtom",
                 4037: "CorePict",
                 4038: "CorePictAtom",
                 4039: "ExPlainAtom",
                 4040: "SrKinsoku",
                 4041: "HandOut",
                 4044: "ExEmbed",
                 4045: "ExEmbedAtom",
                 4046: "ExLink",
                 4047: "ExLinkAtom_old",
                 4048: "BookmarkEntityAtom",
                 4049: "ExLinkAtom",
                 4050: "SrKinsokuAtom",
                 4051: "ExHyperlinkAtom",
                 4053: "ExPlain",
                 4054: "ExPlainLink",
                 4055: "ExHyperlink",
                 4056: "SlideNumberMCAtom",
                 4057: "HeadersFooters",
                 4058: "HeadersFootersAtom",
                 4062: "RecolorEntryAtom",
                 4063: "TxInteractiveInfoAtom",
                 4065: "EmFormatAtom",
                 4066: "CharFormatAtom",
                 4067: "ParaFormatAtom",
                 4068: "MasterText",
                 4071: "RecolorInfoAtom",
                 4073: "ExQuickTime",
                 4074: "ExQuickTimeMovie",
                 4075: "ExQuickTimeMovieData",
                 4076: "ExSubscription",
                 4077: "ExSubscriptionSection",
                 4078: "ExControl",
                 4080: "SlideListWithText",
                 4081: "AnimationInfoAtom",
                 4082: "InteractiveInfo",
                 4083: "InteractiveInfoAtom",
                 4084: "SlideList",
                 4085: "UserEditAtom",
                 4086: "CurrentUserAtom",
                 4087: "DateTimeMCAtom",
                 4088: "GenericDateMCAtom",
                 4090: "FooterMCAtom",
                 4091: "ExControlAtom",
                 4100: "ExMediaAtom",
                 4101: "ExVideo",
                 4102: "ExAviMovie",
                 4103: "ExMCIMovie",
                 4109: "ExMIDIAudio",
                 4110: "ExCDAudio",
                 4111: "ExWAVAudioEmbedded",
                 4112: "ExWAVAudioLink",
                 4113: "ExOleObjStg",
                 4114: "ExCDAudioAtom",
                 4115: "ExWAVAudioEmbeddedAtom",
                 4116: "AnimationInfoAtom",
                 4117: "RTFDateTimeMCAtom",
                 5000: "ProgTags",  # don't know if container
                 5001: "ProgStringTag",
                 5002: "ProgBinaryTag",
                 5003: "BinaryTagData",
                 6000: "PrintOptions",
                 6001: "PersistPtrFullBlock",  # don't know if container
                 6002: "PersistPtrIncrementalBlock",  # don't know if container
                 10000: "RulerIndentAtom",
                 10001: "GScalingAtom",
                 10002: "GRColorAtom",
                 10003: "GLPointAtom",
                 10004: "GlineAtom",
                 11019: "AnimationAtom12",
                 11021: "AnimationHashAtom12",
                 14100: "SlideSyncInfo12",
                 14101: "SlideSyncInfoAtom12",
                 0xf000: "EscherDggContainer",  # Drawing Group Container
                 0xf006: "EscherDgg",
                 0xf016: "EscherCLSID",
                 0xf00b: "EscherOPT",
                 0xf001: "EscherBStoreContainer",
                 0xf007: "EscherBSE",
                 0xf018: "EscherBlip_START",  # Blip types are between
                 0xf117: "EscherBlip_END",  # these two values
                 0xf002: "EscherDgContainer",  # Drawing Container
                 0xf008: "EscherDg",
                 0xf118: "EscherRegroupItems",
                 0xf120: "EscherColorScheme",  # bug in docs
                 0xf003: "EscherSpgrContainer",
                 0xf004: "EscherSpContainer",
                 0xf009: "EscherSpgr",
                 0xf00a: "EscherSp",
                 0xf00c: "EscherTextbox",
                 0xf00d: "EscherClientTextbox",
                 0xf00e: "EscherAnchor",
                 0xf00f: "EscherChildAnchor",
                 0xf010: "EscherClientAnchor",
                 0xf011: "EscherClientData",
                 0xf005: "EscherSolverContainer",
                 0xf012: "EscherConnectorRule",  # bug in docs
                 0xf013: "EscherAlignRule",
                 0xf014: "EscherArcRule",
                 0xf015: "EscherClientRule",
                 0xf017: "EscherCalloutRule",
                 0xf119: "EscherSelection",
                 0xf11a: "EscherColorMRU",
                 0xf11d: "EscherDeletedPspl",  # bug in docs
                 0xf11e: "EscherSplitMenuColors",
                 0xf11f: "EscherOleObject",
                 0xf122: "EscherUserDefined"}

    class CurrentUserAtom(FieldSet):

        def createFields(self):
            yield UInt32(self, "size")
            yield textHandler(UInt32(self, "magic", "0xe391c05f for normal PPT, 0xf3d1c4df for encrypted PPT"), hexadecimal)
            yield UInt32(self, "offsetToCurrentEdit", "Offset in main stream to current edit field")
            yield UInt16(self, "lenUserName", "Length of user name")
            yield UInt16(self, "docFileVersion", "1012 for PP97+")
            yield UInt8(self, "majorVersion", "3 for PP97+")
            yield UInt8(self, "minorVersion", "0 for PP97+")
            yield UInt16(self, "unknown")
            yield String(self, "userName", self["lenUserName"].value, "ANSI version of the username")
            yield UInt32(self, "relVersion", "Release version: 8 for regular PPT file, 9 for multiple-master PPT file")

    class PowerPointObject(FieldSet):

        def createFields(self):
            yield Bits(self, "version", 4)
            yield Bits(self, "instance", 12)
            yield Enum(UInt16(self, "type"), PowerPointDocument.OBJ_TYPES)
            yield UInt32(self, "length")
            self._size = self["length"].value * 8 + 64
            obj_type = self["type"].display
            obj_len = self["length"].value
            # type 1064 (RoundTripCustomTableStyles12) may appear to be a
            # container, but it is not.
            if self["version"].value == 0xF and self["type"].value != 1064:
                while (self.current_size) // 8 < obj_len + 8:
                    yield PowerPointDocument.PowerPointObject(self, "object[]")
            elif obj_len:
                if obj_type == "FontEntityAtom":
                    yield String(self, "data", obj_len, charset="UTF-16-LE", truncate="\0", strip="\0")
                elif obj_type == "TextCharsAtom":
                    yield String(self, "data", obj_len, charset="UTF-16-LE")
                elif obj_type == "TextBytesAtom":
                    yield String(self, "data", obj_len, charset="ASCII")
                elif hasattr(PowerPointDocument, obj_type):
                    field = getattr(PowerPointDocument, obj_type)(self, "data")
                    field._size = obj_len * 8
                    yield field
                else:
                    yield RawBytes(self, "data", obj_len)

        def createDescription(self):
            if self["version"].value == 0xF:
                return "PowerPoint Object Container; instance %s, type %s" % (self["instance"].value, self["type"].display)
            return "PowerPoint Object; version %s, instance %s, type %s" % (self["version"].value, self["instance"].value, self["type"].display)
    ENDIAN_CHECK = False
    OS_CHECK = False

    def createFields(self):
        pos = 0
        while pos // 8 < self.datasize:
            newobj = PowerPointDocument.PowerPointObject(self, "object[]")
            yield newobj
            pos += newobj.size


class CurrentUser(OLE2FragmentParser):

    def createFields(self):
        yield PowerPointDocument.PowerPointObject(self, "current_user")
        if self.current_size < self.size:
            yield String(self, "unicode_name", self["current_user/data/lenUserName"].value * 2, charset="UTF-16-LE")


class ExcelWorkbook(OLE2FragmentParser):
    BIFF_TYPES = {0x000: "DIMENSIONS_v0",
                  0x200: "DIMENSIONS_v2",
                  0x001: "BLANK_v0",
                  0x201: "BLANK_v2",
                  0x002: "INTEGER",
                  0x003: "NUMBER_v0",
                  0x203: "NUMBER_v2",
                  0x004: "LABEL_v0",
                  0x204: "LABEL_v2",
                  0x005: "BOOLERR_v0",
                  0x205: "BOOLERR_v2",
                  0x006: "FORMULA_v0",
                  0x206: "FORMULA_v2",
                  0x406: "FORMULA_v4",
                  0x007: "STRING_v0",
                  0x207: "STRING_v2",
                  0x008: "ROW_v0",
                  0x208: "ROW_v2",
                  0x009: "BOF_v0",
                  0x209: "BOF_v2",
                  0x409: "BOF_v4",
                  0x809: "BOF_v8",
                  0x00a: "EOF",
                  0x00b: "INDEX_v0",
                  0x20b: "INDEX_v2",
                  0x00c: "CALCCOUNT",
                  0x00d: "CALCMODE",
                  0x00e: "PRECISION",
                  0x00f: "REFMODE",
                  0x010: "DELTA",
                  0x011: "ITERATION",
                  0x012: "PROTECT",
                  0x013: "PASSWORD",
                  0x014: "HEADER",
                  0x015: "FOOTER",
                  0x016: "EXTERNCOUNT",
                  0x017: "EXTERNSHEET",
                  0x018: "NAME_v0",
                  0x218: "NAME_v2",
                  0x019: "WINDOWPROTECT",
                  0x01a: "VERTICALPAGEBREAKS",
                  0x01b: "HORIZONTALPAGEBREAKS",
                  0x01c: "NOTE",
                  0x01d: "SELECTION",
                  0x01e: "FORMAT_v0",
                  0x41e: "FORMAT_v4",
                  0x01f: "FORMATCOUNT",  # Undocumented
                  0x020: "COLUMNDEFAULT",  # Undocumented
                  0x021: "ARRAY_v0",
                  0x221: "ARRAY_v2",
                  0x022: "1904",
                  0x023: "EXTERNNAME_v0",
                  0x223: "EXTERNNAME_v2",
                  0x024: "COLWIDTH",  # Undocumented
                  0x025: "DEFAULTROWHEIGHT_v0",
                  0x225: "DEFAULTROWHEIGHT_v2",
                  0x026: "LEFT_MARGIN",
                  0x027: "RIGHT_MARGIN",
                  0x028: "TOP_MARGIN",
                  0x029: "BOTTOM_MARGIN",
                  0x02a: "PRINTHEADERS",
                  0x02b: "PRINTGRIDLINES",
                  0x02f: "FILEPASS",
                  0x031: "FONT_v0",
                  0x231: "FONT_v2",
                  0x032: "FONTCOUNT",  # Undocumented
                  0x033: "PRINTSIZE",  # Undocumented
                  0x036: "TABLE_v0",
                  0x236: "TABLE_v2",
                  0x037: "TABLE2",  # OOo has docs
                  0x038: "WNDESK",  # Undocumented
                  0x039: "ZOOM",  # Undocumented
                  0x03a: "BEGINPREF",  # Undocumented
                  0x03b: "ENDPREF",  # Undocumented
                  0x03c: "CONTINUE",
                  0x03d: "WINDOW1",
                  0x03e: "WINDOW2_v0",
                  0x23e: "WINDOW2_v2",
                  0x03f: "PANE_V2",  # Undocumented
                  0x040: "BACKUP",
                  0x041: "PANE",
                  0x042: "CODEPAGE",
                  0x043: "XF_OLD_v0",
                  0x243: "XF_OLD_v2",
                  0x443: "XF_OLD_v4",
                  0x044: "XF_INDEX",
                  0x045: "FONT_COLOR",
                  0x04d: "PLS",
                  0x050: "DCON",
                  0x051: "DCONREF",
                  0x052: "DCONNAME",
                  0x055: "DEFCOLWIDTH",
                  0x059: "XCT",
                  0x05a: "CRN",
                  0x05b: "FILESHARING",
                  0x05c: "WRITEACCESS",
                  0x05d: "OBJ",
                  0x05e: "UNCALCED",
                  0x05f: "SAVERECALC",
                  0x060: "TEMPLATE",
                  0x061: "INTL",  # Undocumented
                  0x862: "TAB_COLOR",  # Undocumented, OO calls it SHEETLAYOUT
                  0x063: "OBJPROTECT",
                  0x07d: "COLINFO",
                  0x27e: "RK",  # Odd that there is no 0x7e
                  0x07f: "IMDATA",
                  0x080: "GUTS",
                  0x081: "WSBOOL",
                  0x082: "GRIDSET",
                  0x083: "HCENTER",
                  0x084: "VCENTER",
                  0x085: "BOUNDSHEET",
                  0x086: "WRITEPROT",
                  0x087: "ADDIN",
                  0x088: "EDG",
                  0x089: "PUB",
                  0x08c: "COUNTRY",
                  0x08d: "HIDEOBJ",
                  0x08e: "BUNDLESOFFSET",  # Undocumented
                  0x08f: "BUNDLEHEADER",  # Undocumented
                  0x090: "SORT",
                  0x091: "SUB",
                  0x092: "PALETTE",
                  0x293: "STYLE",  # Odd that there is no 0x93
                  0x094: "LHRECORD",
                  0x095: "LHNGRAPH",
                  0x096: "SOUND",
                  0x097: "SYNC",  # Undocumented
                  0x098: "LPR",
                  0x099: "STANDARDWIDTH",
                  0x09a: "FNGROUPNAME",
                  0x09b: "FILTERMODE",
                  0x09c: "FNGROUPCOUNT",
                  0x09d: "AUTOFILTERINFO",
                  0x09e: "AUTOFILTER",
                  0x0a0: "SCL",
                  0x0a1: "SETUP",
                  0x0a4: "TOOLBARVER",  # Undocumented
                  0x0a9: "COORDLIST",
                  0x0ab: "GCW",
                  0x0ae: "SCENMAN",
                  0x0af: "SCENARIO",
                  0x0b0: "SXVIEW",
                  0x0b1: "SXVD",
                  0x0b2: "SXVI",
                  0x0b3: "SXSI",  # Undocumented
                  0x0b4: "SXIVD",
                  0x0b5: "SXLI",
                  0x0b6: "SXPI",
                  0x0b7: "FACENUM",  # Undocumented
                  0x0b8: "DOCROUTE",
                  0x0b9: "RECIPNAME",
                  0x0ba: "SSLIST",  # Undocumented
                  0x0bb: "MASKIMDATA",  # Undocumented
                  0x4bc: "SHRFMLA",
                  0x0bd: "MULRK",
                  0x0be: "MULBLANK",
                  0x0bf: "TOOLBARHDR",  # Undocumented
                  0x0c0: "TOOLBAREND",  # Undocumented
                  0x0c1: "MMS",
                  0x0c2: "ADDMENU",
                  0x0c3: "DELMENU",
                  0x0c4: "TIPHISTORY",  # Undocumented
                  0x0c5: "SXDI",
                  0x0c6: "SXDB",
                  0x0c7: "SXFDB",  # guessed
                  0x0c8: "SXDDB",  # guessed
                  0x0c9: "SXNUM",  # guessed
                  0x0ca: "SXBOOL",  # guessed
                  0x0cb: "SXERR",  # guessed
                  0x0cc: "SXINT",  # guessed
                  0x0cd: "SXSTRING",
                  0x0ce: "SXDTR",  # guessed
                  0x0cf: "SXNIL",  # guessed
                  0x0d0: "SXTBL",
                  0x0d1: "SXTBRGIITM",
                  0x0d2: "SXTBPG",
                  0x0d3: "OBPROJ",
                  0x0d5: "SXIDSTM",
                  0x0d6: "RSTRING",
                  0x0d7: "DBCELL",
                  0x0d8: "SXNUMGROUP",  # from OO : numerical grouping in pivot cache field
                  0x0da: "BOOKBOOL",
                  # 0x0dc: "PARAMQRY",  # DUPLICATE dc
                  0x0dc: "SXEXT",  # DUPLICATE dc
                  0x0dd: "SCENPROTECT",
                  0x0de: "OLESIZE",
                  0x0df: "UDDESC",
                  0x0e0: "XF",
                  0x0e1: "INTERFACEHDR",
                  0x0e2: "INTERFACEEND",
                  0x0e3: "SXVS",
                  0x0e5: "MERGECELLS",  # guessed
                  0x0e9: "BG_PIC",  # Undocumented
                  0x0ea: "TABIDCONF",
                  0x0eb: "MS_O_DRAWING_GROUP",
                  0x0ec: "MS_O_DRAWING",
                  0x0ed: "MS_O_DRAWING_SELECTION",
                  0x0ef: "PHONETIC",  # semi-Undocumented
                  0x0f0: "SXRULE",
                  0x0f1: "SXEX",
                  0x0f2: "SXFILT",
                  0x0f6: "SXNAME",
                  0x0f7: "SXSELECT",
                  0x0f8: "SXPAIR",
                  0x0f9: "SXFMLA",
                  0x0fb: "SXFORMAT",
                  0x0fc: "SST",
                  0x0fd: "LABELSST",
                  0x0ff: "EXTSST",
                  0x100: "SXVDEX",
                  0x103: "SXFORMULA",
                  0x122: "SXDBEX",
                  0x137: "CHTRINSERT",
                  0x138: "CHTRINFO",
                  0x13B: "CHTRCELLCONTENT",
                  0x13d: "TABID",
                  0x140: "CHTRMOVERANGE",
                  0x14D: "CHTRINSERTTAB",
                  0x15F: "LABELRANGES",
                  0x160: "USESELFS",
                  0x161: "DSF",
                  0x162: "XL5MODIFY",
                  0x196: "CHTRHEADER",
                  0x1a5: "FILESHARING2",
                  0x1a9: "USERDBVIEW",
                  0x1aa: "USERSVIEWBEGIN",
                  0x1ab: "USERSVIEWEND",
                  0x1ad: "QSI",
                  0x1ae: "SUPBOOK",
                  0x1af: "PROT4REV",
                  0x1b0: "CONDFMT",
                  0x1b1: "CF",
                  0x1b2: "DVAL",
                  0x1b5: "DCONBIN",
                  0x1b6: "TXO",
                  0x1b7: "REFRESHALL",
                  0x1b8: "HLINK",
                  0x1ba: "CODENAME",  # TYPO in MS Docs
                  0x1bb: "SXFDBTYPE",
                  0x1bc: "PROT4REVPASS",
                  0x1be: "DV",
                  0x1c0: "XL9FILE",
                  0x1c1: "RECALCID",
                  0x800: "LINK_TIP",  # follows an hlink
                  0x802: "UNKNOWN_802",  # OO exports it but has not name or docs
                  0x803: "WQSETT",  # OO named it and can export it, but does not include it in the docs
                  0x804: "WQTABLES",  # OO named it and can export it, but does not include it in the docs
                  # No name or docs, seems related to web query see #153260 for
                  # sample
                  0x805: "UNKNOWN_805",
                  0x810: "PIVOT_AUTOFORMAT",  # Seems to contain pivot table autoformat indicies, plus ??
                  0x864: "UNKNOWN_864",  # seems related to pivot tables
                  0x867: "SHEETPROTECTION",  # OO named it, and has docs
                  0x868: "RANGEPROTECTION",  # OO named it, no docs yet

                  0x1001: "CHART_units",
                  0x1002: "CHART_chart",
                  0x1003: "CHART_series",
                  0x1006: "CHART_dataformat",
                  0x1007: "CHART_lineformat",
                  0x1009: "CHART_markerformat",
                  0x100a: "CHART_areaformat",
                  0x100b: "CHART_pieformat",
                  0x100c: "CHART_attachedlabel",
                  0x100d: "CHART_seriestext",
                  0x1014: "CHART_chartformat",
                  0x1015: "CHART_legend",
                  0x1016: "CHART_serieslist",
                  0x1017: "CHART_bar",
                  0x1018: "CHART_line",
                  0x1019: "CHART_pie",
                  0x101a: "CHART_area",
                  0x101b: "CHART_scatter",
                  0x101c: "CHART_chartline",
                  0x101d: "CHART_axis",
                  0x101e: "CHART_tick",
                  0x101f: "CHART_valuerange",
                  0x1020: "CHART_catserrange",
                  0x1021: "CHART_axislineformat",
                  0x1022: "CHART_chartformatlink",
                  0x1024: "CHART_defaulttext",
                  0x1025: "CHART_text",
                  0x1026: "CHART_fontx",
                  0x1027: "CHART_objectlink",
                  0x1032: "CHART_frame",
                  0x1033: "CHART_begin",
                  0x1034: "CHART_end",
                  0x1035: "CHART_plotarea",
                  0x103a: "CHART_3d",
                  0x103c: "CHART_picf",
                  0x103d: "CHART_dropbar",
                  0x103e: "CHART_radar",
                  0x103f: "CHART_surf",
                  0x1040: "CHART_radararea",
                  0x1041: "CHART_axisparent",
                  0x1043: "CHART_legendxn",
                  0x1044: "CHART_shtprops",
                  0x1045: "CHART_sertocrt",
                  0x1046: "CHART_axesused",
                  0x1048: "CHART_sbaseref",
                  0x104a: "CHART_serparent",
                  0x104b: "CHART_serauxtrend",
                  0x104e: "CHART_ifmt",
                  0x104f: "CHART_pos",
                  0x1050: "CHART_alruns",
                  0x1051: "CHART_ai",
                  0x105b: "CHART_serauxerrbar",
                  0x105c: "CHART_clrtclient",  # Undocumented
                  0x105d: "CHART_serfmt",
                  0x105f: "CHART_3dbarshape",  # Undocumented
                  0x1060: "CHART_fbi",
                  0x1061: "CHART_boppop",
                  0x1062: "CHART_axcext",
                  0x1063: "CHART_dat",
                  0x1064: "CHART_plotgrowth",
                  0x1065: "CHART_siindex",
                  0x1066: "CHART_gelframe",
                  0x1067: "CHART_boppopcustom", }

    class BIFF(FieldSet):

        def createFields(self):
            yield Enum(UInt16(self, "type"), ExcelWorkbook.BIFF_TYPES)
            yield UInt16(self, "length")
            if self["length"].value:
                yield RawBytes(self, "data", self["length"].value)

        def createDescription(self):
            return "Excel BIFF; type %s" % self["type"].display

    def createFields(self):
        pos = 0
        while pos // 8 < self.datasize:
            newobj = ExcelWorkbook.BIFF(self, "BIFF[]")
            yield newobj
            pos += newobj.size


class ThumbsCatalog(OLE2FragmentParser):

    class ThumbsEntry(FieldSet):

        def createFields(self):
            yield UInt32(self, "size")
            yield UInt32(self, "index")
            yield Bits(self, "flags", 8)
            yield RawBytes(self, "unknown[]", 5)
            yield UInt16(self, "unknown[]")
            yield CString(self, "name", charset="UTF-16-LE")
            if self.current_size // 8 != self['size'].value:
                yield RawBytes(self, "padding", self['size'].value - self.current_size // 8)

        def createDescription(self):
            return "Thumbnail entry for %s" % self["name"].display

    def createFields(self):
        yield UInt16(self, "unknown[]")
        yield UInt16(self, "unknown[]")
        yield UInt32(self, "count")
        yield UInt32(self, "unknown[]")
        yield UInt32(self, "unknown[]")
        for i in range(self['count'].value):
            yield ThumbsCatalog.ThumbsEntry(self, "entry[]")


PROPERTY_NAME = {
    "Root Entry": ("root", RootEntry),
    "\5DocumentSummaryInformation": ("doc_summary", DocSummary),
    "\5SummaryInformation": ("summary", Summary),
    "\1CompObj": ("compobj", CompObj),
    "Pictures": ("pictures", Pictures),
    "PowerPoint Document": ("powerpointdoc", PowerPointDocument),
    "Current User": ("current_user", CurrentUser),
    "Workbook": ("workbook", ExcelWorkbook),
    "Catalog": ("catalog", ThumbsCatalog),
    "WordDocument": ("word_doc", WordDocumentParser),
    "0Table": ("table0", WordTableParser),
    "1Table": ("table1", WordTableParser),
}



"""LZX data stream parser.

Also includes a decompression function (slow!!) which can decompress
LZX data stored in a Hachoir stream.

Author: Robert Xiao
Creation date: July 18, 2007
"""
from hachoir.parser import Parser
from hachoir.field import (FieldSet,
                           UInt32, Bit, Bits, PaddingBits,
                           RawBytes, ParserError)
from hachoir.core.endian import MIDDLE_ENDIAN, LITTLE_ENDIAN
from hachoir.core.tools import paddingSize
from hachoir.parser.archive.zlib import build_tree, HuffmanCode, extend_data
import struct


class LZXPreTreeEncodedTree(FieldSet):

    def __init__(self, parent, name, num_elements, *args, **kwargs):
        FieldSet.__init__(self, parent, name, *args, **kwargs)
        self.num_elements = num_elements

    def createFields(self):
        for i in range(20):
            yield Bits(self, "pretree_lengths[]", 4)
        pre_tree = build_tree(
            [self['pretree_lengths[%d]' % x].value for x in range(20)])
        if not hasattr(self.root, "lzx_tree_lengths_" + self.name):
            self.lengths = [0] * self.num_elements
            setattr(self.root, "lzx_tree_lengths_" + self.name, self.lengths)
        else:
            self.lengths = getattr(self.root, "lzx_tree_lengths_" + self.name)
        i = 0
        while i < self.num_elements:
            field = HuffmanCode(self, "tree_code[]", pre_tree)
            if field.realvalue <= 16:
                self.lengths[i] = (self.lengths[i] - field.realvalue) % 17
                field._description = "Literal tree delta length %i (new length value %i for element %i)" % (
                    field.realvalue, self.lengths[i], i)
                i += 1
                yield field
            elif field.realvalue == 17:
                field._description = "Tree Code 17: Zeros for 4-19 elements"
                yield field
                extra = Bits(self, "extra[]", 4)
                zeros = 4 + extra.value
                extra._description = "Extra bits: zeros for %i elements (elements %i through %i)" % (
                    zeros, i, i + zeros - 1)
                yield extra
                self.lengths[i:i + zeros] = [0] * zeros
                i += zeros
            elif field.realvalue == 18:
                field._description = "Tree Code 18: Zeros for 20-51 elements"
                yield field
                extra = Bits(self, "extra[]", 5)
                zeros = 20 + extra.value
                extra._description = "Extra bits: zeros for %i elements (elements %i through %i)" % (
                    zeros, i, i + zeros - 1)
                yield extra
                self.lengths[i:i + zeros] = [0] * zeros
                i += zeros
            elif field.realvalue == 19:
                field._description = "Tree Code 19: Same code for 4-5 elements"
                yield field
                extra = Bits(self, "extra[]", 1)
                run = 4 + extra.value
                extra._description = "Extra bits: run for %i elements (elements %i through %i)" % (
                    run, i, i + run - 1)
                yield extra
                newfield = HuffmanCode(self, "tree_code[]", pre_tree)
                assert newfield.realvalue <= 16
                newfield._description = "Literal tree delta length %i (new length value %i for elements %i through %i)" % (
                    newfield.realvalue, self.lengths[i], i, i + run - 1)
                self.lengths[
                    i:i + run] = [(self.lengths[i] - newfield.realvalue) % 17] * run
                i += run
                yield newfield


class LZXBlock(FieldSet):
    WINDOW_SIZE = {15: 30,
                   16: 32,
                   17: 34,
                   18: 36,
                   19: 38,
                   20: 42,
                   21: 50}
    POSITION_SLOTS = {0: (0, 0, 0),
                      1: (1, 1, 0),
                      2: (2, 2, 0),
                      3: (3, 3, 0),
                      4: (4, 5, 1),
                      5: (6, 7, 1),
                      6: (8, 11, 2),
                      7: (12, 15, 2),
                      8: (16, 23, 3),
                      9: (24, 31, 3),
                      10: (32, 47, 4),
                      11: (48, 63, 4),
                      12: (64, 95, 5),
                      13: (96, 127, 5),
                      14: (128, 191, 6),
                      15: (192, 255, 6),
                      16: (256, 383, 7),
                      17: (384, 511, 7),
                      18: (512, 767, 8),
                      19: (768, 1023, 8),
                      20: (1024, 1535, 9),
                      21: (1536, 2047, 9),
                      22: (2048, 3071, 10),
                      23: (3072, 4095, 10),
                      24: (4096, 6143, 11),
                      25: (6144, 8191, 11),
                      26: (8192, 12287, 12),
                      27: (12288, 16383, 12),
                      28: (16384, 24575, 13),
                      29: (24576, 32767, 13),
                      30: (32768, 49151, 14),
                      31: (49152, 65535, 14),
                      32: (65536, 98303, 15),
                      33: (98304, 131071, 15),
                      34: (131072, 196607, 16),
                      35: (196608, 262143, 16),
                      36: (262144, 393215, 17),
                      37: (393216, 524287, 17),
                      38: (524288, 655359, 17),
                      39: (655360, 786431, 17),
                      40: (786432, 917503, 17),
                      41: (917504, 1048575, 17),
                      42: (1048576, 1179647, 17),
                      43: (1179648, 1310719, 17),
                      44: (1310720, 1441791, 17),
                      45: (1441792, 1572863, 17),
                      46: (1572864, 1703935, 17),
                      47: (1703936, 1835007, 17),
                      48: (1835008, 1966079, 17),
                      49: (1966080, 2097151, 17),
                      }

    def createFields(self):
        yield Bits(self, "block_type", 3)
        yield Bits(self, "block_size", 24)
        self.uncompressed_size = self["block_size"].value
        self.compression_level = self.root.compr_level
        self.window_size = self.WINDOW_SIZE[self.compression_level]
        self.block_type = self["block_type"].value
        curlen = len(self.parent.uncompressed_data)
        intel_started = False  # Do we perform Intel jump fixups on this block?

        if self.block_type in (1, 2):  # Verbatim or aligned offset block
            if self.block_type == 2:
                for i in range(8):
                    yield Bits(self, "aligned_len[]", 3)
                aligned_tree = build_tree(
                    [self['aligned_len[%d]' % i].value for i in range(8)])
            yield LZXPreTreeEncodedTree(self, "main_tree_start", 256)
            yield LZXPreTreeEncodedTree(self, "main_tree_rest", self.window_size * 8)
            main_tree = build_tree(
                self["main_tree_start"].lengths + self["main_tree_rest"].lengths)
            if self["main_tree_start"].lengths[0xE8]:
                intel_started = True
            yield LZXPreTreeEncodedTree(self, "length_tree", 249)
            length_tree = build_tree(self["length_tree"].lengths)
            current_decoded_size = 0
            while current_decoded_size < self.uncompressed_size:
                if (curlen + current_decoded_size) % 32768 == 0 and (curlen + current_decoded_size) != 0:
                    padding = paddingSize(self.address + self.current_size, 16)
                    if padding:
                        yield PaddingBits(self, "padding[]", padding)
                field = HuffmanCode(self, "main_code[]", main_tree)
                if field.realvalue < 256:
                    field._description = "Literal value %r" % chr(
                        field.realvalue)
                    current_decoded_size += 1
                    self.parent.uncompressed_data.append(field.realvalue)
                    yield field
                    continue
                position_header, length_header = divmod(
                    field.realvalue - 256, 8)
                info = self.POSITION_SLOTS[position_header]
                if info[2] == 0:
                    if info[0] == 0:
                        position = self.parent.r0
                        field._description = "Position Slot %i, Position [R0] (%i)" % (
                            position_header, position)
                    elif info[0] == 1:
                        position = self.parent.r1
                        self.parent.r1 = self.parent.r0
                        self.parent.r0 = position
                        field._description = "Position Slot %i, Position [R1] (%i)" % (
                            position_header, position)
                    elif info[0] == 2:
                        position = self.parent.r2
                        self.parent.r2 = self.parent.r0
                        self.parent.r0 = position
                        field._description = "Position Slot %i, Position [R2] (%i)" % (
                            position_header, position)
                    else:
                        position = info[0] - 2
                        self.parent.r2 = self.parent.r1
                        self.parent.r1 = self.parent.r0
                        self.parent.r0 = position
                        field._description = "Position Slot %i, Position %i" % (
                            position_header, position)
                else:
                    field._description = "Position Slot %i, Positions %i to %i" % (
                        position_header, info[0] - 2, info[1] - 2)
                if length_header == 7:
                    field._description += ", Length Values 9 and up"
                    yield field
                    length_field = HuffmanCode(
                        self, "length_code[]", length_tree)
                    length = length_field.realvalue + 9
                    length_field._description = "Length Code %i, total length %i" % (
                        length_field.realvalue, length)
                    yield length_field
                else:
                    field._description += ", Length Value %i (Huffman Code %i)" % (
                        length_header + 2, field.value)
                    yield field
                    length = length_header + 2
                if info[2]:
                    if self.block_type == 1 or info[2] < 3:  # verbatim
                        extrafield = Bits(
                            self, "position_extra[%s" % field.name.split('[')[1], info[2])
                        position = extrafield.value + info[0] - 2
                        extrafield._description = "Position Extra Bits (%i), total position %i" % (
                            extrafield.value, position)
                        yield extrafield
                    else:  # aligned offset
                        position = info[0] - 2
                        if info[2] > 3:
                            extrafield = Bits(
                                self, "position_verbatim[%s" % field.name.split('[')[1], info[2] - 3)
                            position += extrafield.value * 8
                            extrafield._description = "Position Verbatim Bits (%i), added position %i" % (
                                extrafield.value, extrafield.value * 8)
                            yield extrafield
                        if info[2] >= 3:
                            extrafield = HuffmanCode(
                                self, "position_aligned[%s" % field.name.split('[')[1], aligned_tree)
                            position += extrafield.realvalue
                            extrafield._description = "Position Aligned Bits (%i), total position %i" % (
                                extrafield.realvalue, position)
                            yield extrafield
                    self.parent.r2 = self.parent.r1
                    self.parent.r1 = self.parent.r0
                    self.parent.r0 = position
                extend_data(self.parent.uncompressed_data, length, position)
                current_decoded_size += length
        elif self.block_type == 3:  # Uncompressed block
            padding = paddingSize(self.address + self.current_size, 16)
            if padding:
                yield PaddingBits(self, "padding[]", padding)
            else:
                yield PaddingBits(self, "padding[]", 16)
            self.endian = LITTLE_ENDIAN
            intel_started = True  # apparently intel fixup may be needed on uncompressed blocks?
            yield UInt32(self, "r[]", "New value of R0")
            yield UInt32(self, "r[]", "New value of R1")
            yield UInt32(self, "r[]", "New value of R2")
            self.parent.r0 = self["r[0]"].value
            self.parent.r1 = self["r[1]"].value
            self.parent.r2 = self["r[2]"].value
            yield RawBytes(self, "data", self.uncompressed_size)
            self.parent.uncompressed_data += self["data"].value
            if self["block_size"].value % 2:
                yield PaddingBits(self, "padding", 8)
        else:
            raise ParserError("Unknown block type %d!" % self.block_type)

        # Fixup Intel jumps if necessary
        if (
            intel_started
            and self.parent["filesize_indicator"].value
            and self.parent["filesize"].value > 0
        ):
            # Note that we're decoding a block-at-a-time instead of a frame-at-a-time,
            # so we need to handle the frame boundaries carefully.
            filesize = self.parent["filesize"].value
            start_pos = max(0, curlen - 10)  # We may need to correct something from the last block
            end_pos = len(self.parent.uncompressed_data) - 10
            while 1:
                jmp_pos = self.parent.uncompressed_data.find(b"\xE8", start_pos, end_pos)
                if jmp_pos == -1:
                    break
                if (jmp_pos % 32768) >= (32768 - 10):
                    # jumps at the end of frames are not fixed up
                    start_pos = jmp_pos + 1
                    continue
                abs_off, = struct.unpack("<i", self.parent.uncompressed_data[jmp_pos + 1:jmp_pos + 5])
                if -jmp_pos <= abs_off < filesize:
                    if abs_off < 0:
                        rel_off = abs_off + filesize
                    else:
                        rel_off = abs_off - jmp_pos
                    self.parent.uncompressed_data[jmp_pos + 1:jmp_pos + 5] = struct.pack("<i", rel_off)
                start_pos = jmp_pos + 5


class LZXStream(Parser):
    endian = MIDDLE_ENDIAN

    def createFields(self):
        self.uncompressed_data = bytearray()
        self.r0 = 1
        self.r1 = 1
        self.r2 = 1
        yield Bit(self, "filesize_indicator")
        if self["filesize_indicator"].value:
            yield UInt32(self, "filesize")
        while self.current_size < self.size:
            block = LZXBlock(self, "block[]")
            yield block
            if self.size - self.current_size < 16:
                padding = paddingSize(self.address + self.current_size, 16)
                if padding:
                    yield PaddingBits(self, "padding[]", padding)
                break


def lzx_decompress(stream, window_bits):
    data = LZXStream(stream)
    data.compr_level = window_bits
    for _ in data:
        pass
    return data.uncompressed_data



"""
Constant values about endian.
"""

BIG_ENDIAN = "ABCD"
LITTLE_ENDIAN = "DCBA"
MIDDLE_ENDIAN = "BADC"
NETWORK_ENDIAN = BIG_ENDIAN

endian_name = {
    BIG_ENDIAN: "Big endian",
    LITTLE_ENDIAN: "Little endian",
    MIDDLE_ENDIAN: "Middle endian",
}



"""
SWF (Macromedia/Adobe Flash) file parser.

Documentation:

 - Alexis' SWF Reference:
   http://www.m2osw.com/swf_alexref.html
 - http://www.half-serious.com/swf/format/
 - http://www.anotherbigidea.com/javaswf/
 - http://www.gnu.org/software/gnash/

Author: Victor Stinner
Creation date: 29 october 2006
"""

from hachoir.parser import Parser
from hachoir.field import (FieldSet, ParserError,
                           Bit, Bits, UInt8, UInt16, Int32, UInt32, Int64, CString, Enum,
                           Bytes, RawBytes, NullBits, String, SubFile)
from hachoir.core.endian import LITTLE_ENDIAN, BIG_ENDIAN
from hachoir.core.text_handler import textHandler, filesizeHandler
from hachoir.core.tools import paddingSize, humanFrequency
from hachoir.parser.image.common import RGB
from hachoir.parser.image.jpeg import JpegChunk, JpegFile
from hachoir.stream import StringInputStream, ConcatStream
from hachoir.parser.common.deflate import Deflate, has_deflate
from hachoir.parser.container.action_script import parseActionScript, parseABC
import math

# Maximum file size (50 MB)
MAX_FILE_SIZE = 50 * 1024 * 1024

TWIPS = 20


class RECT(FieldSet):
    endian = BIG_ENDIAN

    def createFields(self):
        yield Bits(self, "nbits", 5)
        nbits = self["nbits"].value
        if not nbits:
            raise ParserError("SWF parser: Invalid RECT field size (0)")
        yield Bits(self, "xmin", nbits, "X minimum in twips")
        yield Bits(self, "xmax", nbits, "X maximum in twips")
        yield Bits(self, "ymin", nbits, "Y minimum in twips")
        yield Bits(self, "ymax", nbits, "Y maximum in twips")
        size = paddingSize(self.current_size, 8)
        if size:
            yield NullBits(self, "padding", size)

    def getWidth(self):
        return math.ceil(float(self["xmax"].value) / TWIPS)

    def getHeight(self):
        return math.ceil(float(self["ymax"].value) / TWIPS)

    def createDescription(self):
        return "Rectangle: %ux%u" % (self.getWidth(), self.getHeight())


class FixedFloat16(FieldSet):

    def createFields(self):
        yield UInt8(self, "float_part")
        yield UInt8(self, "int_part")

    def createValue(self):
        return self["int_part"].value + float(self["float_part"].value) / 256


def parseBackgroundColor(parent, size):
    yield RGB(parent, "color")


def bit2hertz(field):
    return humanFrequency(5512.5 * (2 ** field.value))


SOUND_CODEC_MP3 = 2
SOUND_CODEC = {
    0: "RAW",
    1: "ADPCM",
    SOUND_CODEC_MP3: "MP3",
    3: "Uncompressed",
    6: "Nellymoser",
}


class SoundEnvelope(FieldSet):

    def createFields(self):
        yield UInt8(self, "count")
        for index in range(self["count"].value):
            yield UInt32(self, "mark44[]")
            yield UInt16(self, "level0[]")
            yield UInt16(self, "level1[]")


def parseSoundBlock(parent, size):
    # TODO: Be able to get codec... Need to know last sound "def_sound[]" field
    #    if not (...)sound_header:
    #        raise ParserError("Sound block without header")
    if True:  # sound_header == SOUND_CODEC_MP3:
        yield UInt16(parent, "samples")
        yield UInt16(parent, "left")
    size = (parent.size - parent.current_size) // 8
    if size:
        yield RawBytes(parent, "music_data", size)


def parseStartSound(parent, size):
    yield UInt16(parent, "sound_id")
    yield Bit(parent, "has_in_point")
    yield Bit(parent, "has_out_point")
    yield Bit(parent, "has_loops")
    yield Bit(parent, "has_envelope")
    yield Bit(parent, "no_multiple")
    yield Bit(parent, "stop_playback")
    yield NullBits(parent, "reserved", 2)

    if parent["has_in_point"].value:
        yield UInt32(parent, "in_point")
    if parent["has_out_point"].value:
        yield UInt32(parent, "out_point")
    if parent["has_loops"].value:
        yield UInt16(parent, "loop_count")
    if parent["has_envelope"].value:
        yield SoundEnvelope(parent, "envelope")


def parseDefineSound(parent, size):
    yield UInt16(parent, "sound_id")

    yield Bit(parent, "is_stereo")
    yield Bit(parent, "is_16bit")
    yield textHandler(Bits(parent, "rate", 2), bit2hertz)
    yield Enum(Bits(parent, "codec", 4), SOUND_CODEC)

    yield UInt32(parent, "sample_count")

    if parent["codec"].value == SOUND_CODEC_MP3:
        yield UInt16(parent, "len")

    size = (parent.size - parent.current_size) // 8
    if size:
        yield RawBytes(parent, "music_data", size)


def parseSoundHeader(parent, size):
    yield Bit(parent, "playback_is_stereo")
    yield Bit(parent, "playback_is_16bit")
    yield textHandler(Bits(parent, "playback_rate", 2), bit2hertz)
    yield NullBits(parent, "reserved", 4)

    yield Bit(parent, "sound_is_stereo")
    yield Bit(parent, "sound_is_16bit")
    yield textHandler(Bits(parent, "sound_rate", 2), bit2hertz)
    yield Enum(Bits(parent, "codec", 4), SOUND_CODEC)

    yield UInt16(parent, "sample_count")

    if parent["codec"].value == 2:
        yield UInt16(parent, "latency_seek")


class JpegHeader(FieldSet):
    endian = BIG_ENDIAN

    def createFields(self):
        count = 1
        while True:
            chunk = JpegChunk(self, "jpeg_chunk[]")
            yield chunk
            if 1 < count and chunk["type"].value in (JpegChunk.TAG_SOI, JpegChunk.TAG_EOI):
                break
            count += 1


def parseJpeg(parent, size):
    yield UInt16(parent, "char_id", "Character identifier")
    size -= 2

    code = parent["code"].value
    if code != Tag.TAG_BITS:
        if code == Tag.TAG_BITS_JPEG3:
            yield UInt32(parent, "alpha_offset", "Character identifier")
            size -= 4

        addr = parent.absolute_address + parent.current_size + 16
        if parent.stream.readBytes(addr, 2) in (b"\xff\xdb", b"\xff\xd8"):
            header = JpegHeader(parent, "jpeg_header")
            yield header
            hdr_size = header.size // 8
            size -= hdr_size
        else:
            hdr_size = 0

        if code == Tag.TAG_BITS_JPEG3:
            img_size = parent["alpha_offset"].value - hdr_size
        else:
            img_size = size
    else:
        img_size = size
    yield SubFile(parent, "image", img_size, "JPEG picture", parser=JpegFile)
    if code == Tag.TAG_BITS_JPEG3:
        size = (parent.size - parent.current_size) // 8
        yield RawBytes(parent, "alpha", size, "Image data")


def parseVideoFrame(parent, size):
    yield UInt16(parent, "stream_id")
    yield UInt16(parent, "frame_num")
    if 4 < size:
        yield RawBytes(parent, "video_data", size - 4)


class Export(FieldSet):

    def createFields(self):
        yield UInt16(self, "object_id")
        yield CString(self, "name")


def parseExport(parent, size):
    yield UInt16(parent, "count")
    for index in range(parent["count"].value):
        yield Export(parent, "export[]")


def parseProductInfo(parent, size):
    yield Int32(parent, "product_id")
    yield Int32(parent, "edition")
    yield UInt8(parent, "major_version")
    yield UInt8(parent, "minor_version")
    yield Int64(parent, "build_number")
    yield Int64(parent, "compilation_date")


def parseScriptLimits(parent, size):
    yield UInt16(parent, "max_recursion_limit")
    yield UInt16(parent, "timeout_seconds", "Seconds of processing until the SWF is considered 'stuck'")


def parseSymbolClass(parent, size):
    yield UInt16(parent, "count")
    for index in range(parent["count"].value):
        yield UInt16(parent, "symbol_id[]")
        yield CString(parent, "symbol_name[]")


def parseBinaryData(parent, size):
    yield UInt16(parent, "data_id")
    yield UInt32(parent, "reserved")
    if size > 6:
        yield RawBytes(parent, "data", size - 6)


class Tag(FieldSet):
    TAG_BITS = 6
    TAG_BITS_JPEG2 = 32
    TAG_BITS_JPEG3 = 35
    TAG_DO_ABC_DEFINE = 82
    TAG_INFO = {
        # SWF version 1.0
        0: ("end[]", "End", None),
        1: ("show_frame[]", "Show frame", None),
        2: ("def_shape[]", "Define shape", None),
        3: ("free_char[]", "Free character", None),
        4: ("place_obj[]", "Place object", None),
        5: ("remove_obj[]", "Remove object", None),
        6: ("def_bits[]", "Define bits", parseJpeg),
        7: ("def_but[]", "Define button", None),
        8: ("jpg_table", "JPEG tables", None),
        9: ("bkgd_color[]", "Set background color", parseBackgroundColor),
        10: ("def_font[]", "Define font", None),
        11: ("def_text[]", "Define text", None),
        12: ("action[]", "Action script", parseActionScript),
        13: ("def_font_info[]", "Define font info", None),

        # SWF version 2.0
        14: ("def_sound[]", "Define sound", parseDefineSound),
        15: ("start_sound[]", "Start sound", parseStartSound),
        16: ("stop_sound[]", "Stop sound", None),
        17: ("def_but_sound[]", "Define button sound", None),
        18: ("sound_hdr", "Sound stream header", parseSoundHeader),
        19: ("sound_blk[]", "Sound stream block", parseSoundBlock),
        20: ("def_bits_lossless[]", "Define bits lossless", None),
        21: ("def_bits_jpeg2[]", "Define bits JPEG 2", parseJpeg),
        22: ("def_shape2[]", "Define shape 2", None),
        23: ("def_but_cxform[]", "Define button CXFORM", None),
        24: ("protect", "File is protected", None),

        # SWF version 3.0
        25: ("path_are_ps[]", "Paths are Postscript", None),
        26: ("place_obj2[]", "Place object 2", None),
        28: ("remove_obj2[]", "Remove object 2", None),
        29: ("sync_frame[]", "Synchronize frame", None),
        31: ("free_all[]", "Free all", None),
        32: ("def_shape3[]", "Define shape 3", None),
        33: ("def_text2[]", "Define text 2", None),
        34: ("def_but2[]", "Define button2", None),
        35: ("def_bits_jpeg3[]", "Define bits JPEG 3", parseJpeg),
        36: ("def_bits_lossless2[]", "Define bits lossless 2", None),
        39: ("def_sprite[]", "Define sprite", None),
        40: ("name_character[]", "Name character", None),
        41: ("product_info", "Generator product info", parseProductInfo),
        42: ("generator_text[]", "Generator text", None),
        43: ("frame_label[]", "Frame label", None),
        45: ("sound_hdr2[]", "Sound stream header2", parseSoundHeader),
        46: ("def_morph_shape[]", "Define morph shape", None),
        47: ("gen_frame[]", "Generate frame", None),
        48: ("def_font2[]", "Define font 2", None),
        49: ("tpl_command[]", "Template command", None),

        # SWF version 4.0
        37: ("def_text_field[]", "Define text field", None),
        38: ("def_quicktime_movie[]", "Define QuickTime movie", None),

        # SWF version 5.0
        50: ("def_cmd_obj[]", "Define command object", None),
        51: ("flash_generator", "Flash generator", None),
        52: ("gen_ext_font[]", "Gen external font", None),
        56: ("export[]", "Export", parseExport),
        57: ("import[]", "Import", None),
        58: ("ebnable_debug", "Enable debug", None),

        # SWF version 6.0
        59: ("do_init_action[]", "Do init action", None),
        60: ("video_str[]", "Video stream", None),
        61: ("video_frame[]", "Video frame", parseVideoFrame),
        62: ("def_font_info2[]", "Define font info 2", None),
        63: ("mx4[]", "MX4", None),
        64: ("enable_debug2", "Enable debugger 2", None),

        # SWF version 7.0
        65: ("script_limits[]", "Script limits", parseScriptLimits),
        66: ("tab_index[]", "Set tab index", None),

        # SWF version 8.0
        69: ("file_attr[]", "File attributes", None),
        70: ("place_obj3[]", "Place object 3", None),
        71: ("import2[]", "Import a definition list from another movie", None),
        73: ("def_font_align[]", "Define font alignment zones", None),
        74: ("csm_txt_set[]", "CSM text settings", None),
        75: ("def_font3[]", "Define font text 3", None),
        77: ("metadata[]", "XML code describing the movie", None),
        78: ("def_scale_grid[]", "Define scaling factors", None),
        83: ("def_shape4[]", "Define shape 4", None),
        84: ("def_morph2[]", "Define a morphing shape 2", None),

        # SWF version 9.0
        72: ("do_abc[]", "SWF 9 ActionScript container; actions only", parseABC),
        76: ("symbol_class[]", "Instantiate objects from a set of classes", parseSymbolClass),
        82: ("do_abc_define[]", "SWF 9 ActionScript container; identifier, name, actions", parseABC),
        86: ("def_scene_frame[]", "Define raw data for scenes and frames", None),
        87: ("def_binary_data[]", "Defines a buffer of any size with any binary user data", parseBinaryData),
        88: ("def_font_name[]", "Define the legal font name and copyright", None),
    }

    def __init__(self, *args):
        FieldSet.__init__(self, *args)
        size = self["length"].value
        if self[0].name == "length_ext":
            self._size = (6 + size) * 8
        else:
            self._size = (2 + size) * 8
        code = self["code"].value
        if code in self.TAG_INFO:
            self._name, self._description, self.parser = self.TAG_INFO[code]
        else:
            self.parser = None

    def createFields(self):
        if self.stream.readBits(self.absolute_address, 6, self.endian) == 63:
            yield Bits(self, "length_ext", 6)
            yield Bits(self, "code", 10)
            yield filesizeHandler(UInt32(self, "length"))
        else:
            yield filesizeHandler(Bits(self, "length", 6))
            yield Bits(self, "code", 10)
        size = self["length"].value
        if 0 < size:
            if self.parser:
                yield from self.parser(self, size)
            else:
                yield RawBytes(self, "data", size)

    def createDescription(self):
        return "Tag: %s (%s)" % (self["code"].display, self["length"].display)


class SwfFile(Parser):
    VALID_VERSIONS = set(range(1, 10 + 1))
    PARSER_TAGS = {
        "id": "swf",
        "category": "container",
        "file_ext": ["swf"],
        "mime": ("application/x-shockwave-flash",),
        "min_size": 64,
        "description": "Macromedia Flash data"
    }
    PARSER_TAGS["magic"] = []
    for version in VALID_VERSIONS:
        PARSER_TAGS["magic"].append((b"FWS" + bytes([version]), 0))
        PARSER_TAGS["magic"].append((b"CWS" + bytes([version]), 0))
    endian = LITTLE_ENDIAN
    SWF_SCALE_FACTOR = 1.0 / 20

    def validate(self):
        if self.stream.readBytes(0, 3) not in (b"FWS", b"CWS"):
            return "Wrong file signature"
        if self["version"].value not in self.VALID_VERSIONS:
            return "Unknown version"
        if MAX_FILE_SIZE < self["filesize"].value:
            return "File too big (%u)" % self["filesize"].value
        if self["signature"].value == "FWS":
            if self["rect/padding"].value != 0:
                return "Unknown rectangle padding value"
        return True

    def createFields(self):
        yield String(self, "signature", 3, "SWF format signature", charset="ASCII")
        yield UInt8(self, "version")
        yield filesizeHandler(UInt32(self, "filesize"))
        if self["signature"].value != "CWS":
            yield RECT(self, "rect")
            yield FixedFloat16(self, "frame_rate")
            yield UInt16(self, "frame_count")

            while not self.eof:
                yield Tag(self, "tag[]")
        else:
            size = (self.size - self.current_size) // 8
            if has_deflate:
                data = Deflate(Bytes(self, "compressed_data", size), False)

                def createInputStream(cis, source=None, **args):
                    stream = cis(source=source)
                    header = StringInputStream(
                        b"FWS" + self.stream.readBytes(3 * 8, 5))
                    args.setdefault("tags", []).append(("class", SwfFile))
                    return ConcatStream((header, stream), source=stream.source, **args)
                data.setSubIStream(createInputStream)
                yield data
            else:
                yield Bytes(self, "compressed_data", size)

    def createDescription(self):
        desc = ["version %u" % self["version"].value]
        if self["signature"].value == "CWS":
            desc.append("compressed")
        return "Macromedia Flash data: %s" % (", ".join(desc))

    def createContentSize(self):
        if self["signature"].value == "FWS":
            return self["filesize"].value * 8
        else:
            # TODO: Size of compressed Flash?
            return None



from hachoir.parser.video.asf import AsfFile  # noqa
from hachoir.parser.video.flv import FlvFile  # noqa
from hachoir.parser.video.mpeg_video import MPEGVideoFile  # noqa
from hachoir.parser.video.mpeg_ts import MPEG_TS  # noqa



from hachoir.field import Field, FieldError
from hachoir.stream import InputStream
from hachoir.core.endian import BIG_ENDIAN, LITTLE_ENDIAN, MIDDLE_ENDIAN
from hachoir.core.event_handler import EventHandler


class ParserError(FieldError):
    """
    Error raised by a field set.

    @see: L{FieldError}
    """
    pass


class MatchError(FieldError):
    """
    Error raised by a field set when the stream content doesn't
    match to file format.

    @see: L{FieldError}
    """
    pass


class BasicFieldSet(Field):
    _event_handler = None
    is_field_set = True
    endian = None

    def __init__(self, parent, name, stream, description, size):
        # Sanity checks (preconditions)
        assert not parent or issubclass(parent.__class__, BasicFieldSet)
        assert issubclass(stream.__class__, InputStream)

        # Set field set size
        if size is None and self.static_size:
            assert isinstance(self.static_size, int)
            size = self.static_size

        # Set Field attributes
        self._parent = parent
        self._name = name
        self._size = size
        self._description = description
        self.stream = stream
        self._field_array_count = {}

        # Set endian
        if not self.endian:
            assert parent and parent.endian
            self.endian = parent.endian

        if parent:
            # This field set is one of the root leafs
            self._address = parent.nextFieldAddress()
            self.root = parent.root
            assert id(self.stream) == id(parent.stream)
        else:
            # This field set is the root
            self._address = 0
            self.root = self
            self._global_event_handler = None

        # Sanity checks (post-conditions)
        assert self.endian in (BIG_ENDIAN, LITTLE_ENDIAN, MIDDLE_ENDIAN)
        if (self._size is not None) and (self._size <= 0):
            raise ParserError("Invalid parser '%s' size: %s" %
                              (self.path, self._size))

    def reset(self):
        self._field_array_count = {}

    def createValue(self):
        return None

    def connectEvent(self, event_name, handler, local=True):
        assert event_name in (
            # Callback prototype: def f(field)
            # Called when new value is already set
            "field-value-changed",

            # Callback prototype: def f(field)
            # Called when field size is already set
            "field-resized",

            # A new field has been inserted in the field set
            # Callback prototype: def f(index, new_field)
            "field-inserted",

            # Callback prototype: def f(old_field, new_field)
            # Called when new field is already in field set
            "field-replaced",

            # Callback prototype: def f(field, new_value)
            # Called to ask to set new value
            "set-field-value"
        ), "Event name %r is invalid" % event_name
        if local:
            if self._event_handler is None:
                self._event_handler = EventHandler()
            self._event_handler.connect(event_name, handler)
        else:
            if self.root._global_event_handler is None:
                self.root._global_event_handler = EventHandler()
            self.root._global_event_handler.connect(event_name, handler)

    def raiseEvent(self, event_name, *args):
        # Transfer event to local listeners
        if self._event_handler is not None:
            self._event_handler.raiseEvent(event_name, *args)

        # Transfer event to global listeners
        if self.root._global_event_handler is not None:
            self.root._global_event_handler.raiseEvent(event_name, *args)

    def setUniqueFieldName(self, field):
        key = field._name[:-2]
        try:
            self._field_array_count[key] += 1
        except KeyError:
            self._field_array_count[key] = 0
        field._name = key + "[%u]" % self._field_array_count[key]

    def readFirstFields(self, number):
        """
        Read first number fields if they are not read yet.

        Returns number of new added fields.
        """
        number = number - self.current_length
        if 0 < number:
            return self.readMoreFields(number)
        else:
            return 0

    def createFields(self):
        raise NotImplementedError()

    def __iter__(self):
        raise NotImplementedError()

    def __len__(self):
        raise NotImplementedError()

    def getField(self, key, const=True):
        raise NotImplementedError()

    def nextFieldAddress(self):
        raise NotImplementedError()

    def getFieldIndex(self, field):
        raise NotImplementedError()

    def readMoreFields(self, number):
        raise NotImplementedError()



import os
import sys
import time
import hachoir.core.config as config


class Log:
    LOG_INFO = 0
    LOG_WARN = 1
    LOG_ERROR = 2

    level_name = {
        LOG_WARN: "[warn]",
        LOG_ERROR: "[err!]",
        LOG_INFO: "[info]"
    }

    def __init__(self):
        self.__buffer = {}
        self.__file = None
        self.use_print = True
        self.use_buffer = False
        # Prototype: def func(level, prefix, text, context)
        self.on_new_message = None

    def shutdown(self):
        if self.__file:
            self._writeIntoFile("Stop Hachoir")

    def setFilename(self, filename, append=True):
        """
        Use a file to store all messages. The
        UTF-8 encoding will be used. Write an informative
        message if the file can't be created.

        @param filename: C{L{string}}
        """

        # Look if file already exists or not
        filename = os.path.expanduser(filename)
        filename = os.path.realpath(filename)
        append = os.access(filename, os.F_OK)

        # Create log file (or open it in append mode, if it already exists)
        try:
            import codecs
            if append:
                self.__file = codecs.open(filename, "a", "utf-8")
            else:
                self.__file = codecs.open(filename, "w", "utf-8")
            self._writeIntoFile("Starting Hachoir")
        except FileNotFoundError:
            self.__file = None
            self.info("[Log] setFilename(%s) fails: no such file"
                      % filename)

    def _writeIntoFile(self, message):
        timestamp = time.strftime("%Y-%m-%d %H:%M:%S")
        self.__file.write("%s - %s\n" % (timestamp, message))
        self.__file.flush()

    def newMessage(self, level, text, ctxt=None):
        """
        Write a new message : append it in the buffer,
        display it to the screen (if needed), and write
        it in the log file (if needed).

        @param level: Message level.
        @type level: C{int}
        @param text: Message content.
        @type text: C{str}
        @param ctxt: The caller instance.
        """

        if level < self.LOG_ERROR and config.quiet or \
           level <= self.LOG_INFO and not config.verbose:
            return
        if config.debug:
            from hachoir.core.error import getBacktrace
            backtrace = getBacktrace(None)
            if backtrace:
                text += "\n\n" + backtrace

        _text = text
        if hasattr(ctxt, "_logger"):
            _ctxt = ctxt._logger()
            if _ctxt is not None:
                text = "[%s] %s" % (_ctxt, text)

        # Add message to log buffer
        if self.use_buffer:
            if level not in self.__buffer:
                self.__buffer[level] = [text]
            else:
                self.__buffer[level].append(text)

        # Add prefix
        prefix = self.level_name.get(level, "[info]")

        # Display on stdout (if used)
        if self.use_print:
            sys.stdout.flush()
            sys.stderr.write("%s %s\n" % (prefix, text))
            sys.stderr.flush()

        # Write into outfile (if used)
        if self.__file:
            self._writeIntoFile("%s %s" % (prefix, text))

        # Use callback (if used)
        if self.on_new_message:
            self.on_new_message(level, prefix, _text, ctxt)

    def info(self, text):
        """
        New informative message.
        @type text: C{str}
        """
        self.newMessage(Log.LOG_INFO, text)

    def warning(self, text):
        """
        New warning message.
        @type text: C{str}
        """
        self.newMessage(Log.LOG_WARN, text)

    def error(self, text):
        """
        New error message.
        @type text: C{str}
        """
        self.newMessage(Log.LOG_ERROR, text)


log = Log()


class Logger(object):

    def _logger(self):
        return "<%s>" % self.__class__.__name__

    def info(self, text):
        log.newMessage(Log.LOG_INFO, text, self)

    def warning(self, text):
        log.newMessage(Log.LOG_WARN, text, self)

    def error(self, text):
        log.newMessage(Log.LOG_ERROR, text, self)
